import fsui
from launcher.i18n import gettext
from launcher.settings.option_ui import OptionUI


class OpenGLSettingsPage(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        label = fsui.HeadingLabel(self, gettext("OpenGL Settings"))
        self.layout.add(label, margin=10, margin_bottom=20)

        def add_option(name):
            self.layout.add(
                OptionUI.create_group(self, name),
                fill=True,
                margin_top=10,
                margin_bottom=10,
            )

        add_option("fsaa")
        add_option("texture_filter")
        add_option("video_sync_method")
        add_option("texture_format")



from __future__ import absolute_import
from __future__ import print_function

import os
import os.path
import sys
import unicodedata

from .ADFSDir import ADFSDir
from .ADFSFile import ADFSFile
from .ADFSVolume import ADFSVolume
from .MetaDB import MetaDB
from . import DosType
from amitools.fs.block.BootBlock import BootBlock
from amitools.fs.blkdev.BlkDevFactory import BlkDevFactory
from amitools.fs.blkdev.DiskGeometry import DiskGeometry
import amitools.util.KeyValue as KeyValue
from .FSString import FSString

class Imager:
  def __init__(self, path_encoding=None):
    self.meta_db = None
    self.total_bytes = 0
    self.path_encoding = path_encoding
    # get path name encoding for host file system
    if self.path_encoding == None:
      self.path_encoding = sys.getfilesystemencoding()

  def get_total_bytes(self):
    return self.total_bytes

  def to_path_str(self, u):
    """convert a unicode string to OS path name encoding"""
    if type(u) != unicode:
      raise ValueError("to_path_str: must pass a unicode string")
    return u.encode(self.path_encoding)

  def from_path_str(self, s):
    """convert a OS path name encoded string to unicode"""
    if type(s) != str:
      raise ValueError("from_path_str: must pass a string")
    u = s.decode(self.path_encoding)
    # on Mac OS X normalize from decomposed form
    if sys.platform.startswith('darwin'):
      return unicodedata.normalize('NFC',u)
    else:
      return u

  # ----- unpack -----

  def unpack(self, volume, out_path):
    # check for volume path
    vol_name = volume.name.get_unicode()
    if not os.path.exists(out_path):
      vol_path = out_path
    else:
      path = os.path.abspath(out_path)
      vol_path = os.path.join(path, self.to_path_str(vol_name))
    if os.path.exists(vol_path):
      raise IOError("Unpack directory already exists: "+vol_path)
    # check for meta file
    meta_path = vol_path + ".xdfmeta"
    if os.path.exists(meta_path):
      raise IOError("Unpack meta file already exists:"+meta_path)
    # check for block dev file
    blkdev_path = vol_path + ".blkdev"
    if os.path.exists(blkdev_path):
      raise IOError("Unpack blkdev file aready exists:"+blkdev_path)
    # create volume path
    self.meta_db = MetaDB()
    self.unpack_root(volume, vol_path)
    # save meta db
    self.meta_db.set_volume_name(volume.name.get_unicode())
    self.meta_db.set_root_meta_info(volume.get_meta_info())
    self.meta_db.set_dos_type(volume.boot.dos_type)
    self.meta_db.save(meta_path)
    # save boot code
    if volume.boot.boot_code != None:
      boot_code_path = vol_path + ".bootcode"
      f = open(boot_code_path,"wb")
      f.write(volume.boot.boot_code)
      f.close()
    # save blkdev
    f = open(blkdev_path,"wb")
    f.write("%s\n" % volume.blkdev.get_chs_str())
    f.close()

  def unpack_root(self, volume, vol_path):
    self.unpack_dir(volume.get_root_dir(), vol_path)

  def unpack_dir(self, dir, path):
    if not os.path.exists(path):
      os.mkdir(path)
    for e in dir.get_entries():
      self.unpack_node(e, path)

  def unpack_node(self, node, path):
    name = node.name.get_unicode_name()
    # store meta info
    if self.meta_db != None:
      # get path as FSString
      node_path = node.get_node_path_name()
      self.meta_db.set_meta_info(node_path.get_unicode(), node.meta_info)
    # sub dir
    if node.is_dir():
      sub_dir = os.path.join(path, self.to_path_str(name))
      os.mkdir(sub_dir)
      for sub_node in node.get_entries():
        self.unpack_node(sub_node, sub_dir)
      node.flush()
    # file
    elif node.is_file():
      data = node.get_file_data()
      node.flush()
      file_path = os.path.join(path, self.to_path_str(name))
      fh = open(file_path, "wb")
      fh.write(data)
      fh.close()
      self.total_bytes += len(data)

  # ----- pack -----

  def pack(self, in_path, image_file, force=True, options=None, dos_type=None):
    self.pack_begin(in_path)
    blkdev = self.pack_create_blkdev(in_path, image_file, force, options)
    if blkdev == None:
      raise IOError("Can't create block device for image: "+in_path)
    volume = self.pack_create_volume(in_path, blkdev, dos_type)
    if not volume.valid:
      raise IOError("Can't create volume for image: "+in_path)
    self.pack_root(in_path, volume)
    self.pack_end(in_path, volume)

  def pack_begin(self, in_path):
    # remove trailing slash
    if in_path[-1] == '/':
      in_path = in_path[:-1]
    meta_path = in_path + ".xdfmeta"
    if os.path.exists(meta_path):
      self.meta_db = MetaDB()
      self.meta_db.load(meta_path)

  def pack_end(self, in_path, volume):
    boot_code_path = in_path + ".bootcode"
    if os.path.exists(boot_code_path):
      # read boot code
      f = open(boot_code_path, "rb")
      data = f.read()
      f.close()
      # write boot code
      bb = volume.boot
      ok = bb.set_boot_code(data)
      if ok:
        bb.write()
      else:
        raise IOError("Invalid Boot Code")

  def pack_create_blkdev(self, in_path, image_file, force=True, options=None):
    # try to read options from blkdev file
    if options == None or len(options) == 0:
      blkdev_path = in_path + ".blkdev"
      if os.path.exists(blkdev_path):
        f = open(blkdev_path, "rb")
        options = {}
        for line in f:
          KeyValue.parse_key_value_string(line, options)
        f.close()
    f = BlkDevFactory()
    return f.create(image_file, force=force, options=options)

  def pack_create_volume(self, in_path, blkdev, dos_type=None):
    if self.meta_db != None:
      name = self.meta_db.get_volume_name()
      meta_info = self.meta_db.get_root_meta_info()
      if dos_type is None:
        dos_type = self.meta_db.get_dos_type()
    else:
      # try to derive volume name from image name
      if in_path == None or in_path == "":
        raise IOError("Invalid pack input path!")
      # remove trailing slash
      if in_path[-1] == '/':
        in_path = in_path[:-1]
      name = self.from_path_str(os.path.basename(in_path))
      meta_info = None
      if dos_type is None:
        dos_type = DosType.DOS0
    volume = ADFSVolume(blkdev)
    volume.create(FSString(name), meta_info, dos_type=dos_type)
    return volume

  def pack_root(self, in_path, volume):
    self.pack_dir(in_path, volume.get_root_dir())

  def pack_dir(self, in_path, parent_node):
    path = os.path.abspath(in_path)
    if not os.path.exists(path):
      raise IOError("Pack directory does not exist: "+path)
    for name in os.listdir(in_path):
      sub_path = os.path.join(in_path, name)
      self.pack_entry(sub_path, parent_node)

  def pack_entry(self, in_path, parent_node):
    ami_name = self.from_path_str(os.path.basename(in_path))
    # retrieve meta info for path from DB
    if self.meta_db != None:
      ami_path = parent_node.get_node_path_name().get_unicode()
      if ami_path != u"":
        ami_path += u"/" + ami_name
      else:
        ami_path = ami_name
      meta_info = self.meta_db.get_meta_info(ami_path)
    else:
      meta_info = None

    # pack directory
    if os.path.isdir(in_path):
      node = parent_node.create_dir(FSString(ami_name), meta_info, False)
      for name in os.listdir(in_path):
        sub_path = os.path.join(in_path, name)
        self.pack_entry(sub_path, node)
      node.flush()
    # pack file
    elif os.path.isfile(in_path):
      # read file
      fh = open(in_path, "rb")
      data = fh.read()
      fh.close()
      node = parent_node.create_file(FSString(ami_name), data, meta_info, False)
      node.flush()
      self.total_bytes += len(data)



import functools
import sys
from typing import Any

try:
    import PyQt6  # type: ignore # noqa: F401
except ImportError:
    # noinspection PyUnresolvedReferences, PyPackageRequirements
    from PyQt5.QtCore import *  # type: ignore # noqa: F403
    from PyQt5.QtCore import Qt, QUrl  # type: ignore

    # noinspection PyUnresolvedReferences, PyPackageRequirements
    from PyQt5.QtCore import pyqtSignal as QSignal  # type: ignore # noqa: F401

    # noinspection PyUnresolvedReferences, PyPackageRequirements
    from PyQt5.QtCore import pyqtSignal as Signal  # type: ignore # noqa: F401

    # noinspection PyUnresolvedReferences, PyPackageRequirements
    from PyQt5.QtGui import *  # type: ignore # noqa: F403
    from PyQt5.QtGui import QColor, QDesktopServices, QFont, QPalette  # type: ignore

    # noinspection PyUnresolvedReferences, PyPackageRequirements
    from PyQt5.QtOpenGL import *  # type: ignore # noqa: F403

    # noinspection PyUnresolvedReferences, PyPackageRequirements
    from PyQt5.QtWidgets import *  # type: ignore # noqa: F403
    from PyQt5.QtWidgets import QApplication, QStyleFactory  # type: ignore

    try:
        # noinspection PyUnresolvedReferences, PyPackageRequirements
        from PyQt5.QtX11Extras import *  # type: ignore # noqa: F403
    except ImportError:
        pass

    def qflags(obj: Any) -> int:
        return int(obj)

    def qvalue(obj: Any) -> int:
        return obj

    is_pyqt6 = False

else:
    # noinspection PyUnresolvedReferences, PyPackageRequirements
    from PyQt6.QtCore import *  # type: ignore # noqa: F403
    from PyQt6.QtCore import Qt, QUrl

    # noinspection PyUnresolvedReferences, PyPackageRequirements
    from PyQt6.QtCore import pyqtSignal as QSignal  # noqa: F401

    # noinspection PyUnresolvedReferences, PyPackageRequirements
    from PyQt6.QtCore import pyqtSignal as Signal  # noqa: F401

    # noinspection PyUnresolvedReferences, PyPackageRequirements
    from PyQt6.QtGui import *  # type: ignore # noqa: F403
    from PyQt6.QtGui import QColor, QDesktopServices, QFont, QPalette

    # noinspection PyUnresolvedReferences, PyPackageRequirements
    from PyQt6.QtOpenGL import *  # type: ignore # noqa: F403

    # noinspection PyUnresolvedReferences, PyPackageRequirements
    from PyQt6.QtWidgets import *  # type: ignore # noqa: F403
    from PyQt6.QtWidgets import QApplication, QStyleFactory

    try:
        # noinspection PyUnresolvedReferences, PyPackageRequirements
        from PyQt6.QtX11Extras import *  # type: ignore # noqa: F403
    except ImportError:
        pass

    def qflags(obj: Any) -> int:
        return int(obj.value)

    def qvalue(obj: Any) -> int:
        return obj.value

    is_pyqt6 = True


import fsbc.application
import fsbc.desktop
import fstd.desktop
from fsbc.settings import Settings
from fsgs.option import Option


def open_url_in_browser(url):
    print("[QT] open_url_in_browser", url)
    # noinspection PyCallByClass,PyTypeChecker,PyArgumentList
    QDesktopServices.openUrl(QUrl(url))


# class QtBaseApplication(QApplication):
#     pass


@functools.lru_cache()
def init_qt():
    # if sys.platform == "darwin":
    #     # qt_plugins_dir = os.path.join(
    #     #     BaseApplication.executable_dir(), "..", "Resources",
    #     #     "qt_plugins")
    #     # print(qt_plugins_dir)
    #     # if os.path.exists(qt_plugins_dir):
    #     #     QApplication.setLibraryPaths([qt_plugins_dir])
    #     if os.path.exists(
    #         os.path.join(
    #             fsboot.executable_dir(), "platforms", "libqcocoa.dylib"
    #         )
    #     ):
    #         # noinspection PyCallByClass,PyTypeChecker,PyArgumentList
    #         QApplication.setLibraryPaths([fsboot.executable_dir()])

    # if getattr(sys, "frozen", False):
    #     # noinspection PyCallByClass,PyTypeChecker,PyArgumentList
    #     QApplication.setLibraryPaths([fsboot.executable_dir()])

    # Should not be necessary with Qt 5.2.x:
    # fix_qt_for_maverick()

    # QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)

    fsbc.desktop.set_open_url_in_browser_function(open_url_in_browser)
    qapplication = QApplication(sys.argv)
    initialize_qt_style(qapplication)
    return qapplication


# noinspection PyUnboundLocalVariable
def initialize_qt_style(qapplication):
    # from fsui.qt import QStyleFactory, QPalette, QColor, Qt, QFont
    fusion_variant = ""

    pa = QPalette()

    launcher_theme = Settings.instance().get(Option.LAUNCHER_THEME)
    if launcher_theme == "standard":
        use_fusion_theme = False
    elif launcher_theme == "native":
        # native is an older deprecated name for standard
        use_fusion_theme = False
    elif launcher_theme == "fusion-adwaita":
        use_fusion_theme = True
        fusion_variant = "adwaita"
    elif launcher_theme == "fusion-plain":
        use_fusion_theme = True
    elif launcher_theme == "fusion-dark":
        use_fusion_theme = True
        fusion_variant = "dark"
    elif launcher_theme == "fusion-windows10":
        use_fusion_theme = True
        fusion_variant = "windows10"
    elif launcher_theme == "fws":
        use_fusion_theme = True
        fusion_variant = "fws"
        from fsui.qt.window import FwsWindow

        FwsWindow.set_default()
    else:
        use_fusion_theme = True

        qt_color_scheme_dark = None
        try:
            qt_color_scheme_dark = qapplication.styleHints().colorScheme() == Qt.ColorScheme.Dark
        except Exception:
            # It only works on Qt 6.5 or higher
            pass

        if qt_color_scheme_dark or pa.color(QPalette.ColorRole.Window).red() < 128:
            fusion_variant = "dark"
        elif fstd.desktop.is_running_gnome_3():
            fusion_variant = "adwaita"
        elif fstd.desktop.is_running_windows_10():
            fusion_variant = "windows10"

    if "--launcher-theme=fusion-dark" in sys.argv:
        use_fusion_theme = True
        fusion_variant = "dark"

    font = qapplication.font()
    print("FONT: Default is {} {}".format(font.family(), font.pointSize()))

    if use_fusion_theme:
        # noinspection PyCallByClass,PyTypeChecker,PyArgumentList
        qapplication.setStyle(QStyleFactory.create("Fusion"))

        if fusion_variant == "adwaita":
            background = QColor("#f2f2f2")
            pa.setColor(QPalette.ColorRole.Window, background)
            pa.setColor(QPalette.ColorRole.AlternateBase, background)
            pa.setColor(QPalette.ColorRole.Button, background)
            # pa.setColor(QPalette.ColorRole.Base, QColor(255, 255, 255))
            pa.setColor(
                QPalette.ColorGroup.Disabled,
                QPalette.ColorRole.Base,
                QColor(241, 241, 241),
            )

            # pa.setColor(QPalette.ColorRole.Window, QColor("#aeaeae"))
            # pa.setColor(QPalette.ColorRole.AlternateBase, QColor("#aeaeae"))
            # pa.setColor(QPalette.ColorRole.Button, QColor("#aeaeae"))

        elif fusion_variant == "fws" or fusion_variant == "windows10":
            pa.setColor(QPalette.ColorRole.Window, QColor(242, 242, 242))
            pa.setColor(
                QPalette.ColorRole.AlternateBase, QColor(242, 242, 242)
            )
            pa.setColor(QPalette.ColorRole.Button, QColor(242, 242, 242))
        elif fusion_variant == "dark":
            pa.setColor(QPalette.ColorRole.Window, QColor(53, 53, 53))
            pa.setColor(QPalette.ColorRole.WindowText, Qt.GlobalColor.white)
            pa.setColor(QPalette.ColorRole.Base, QColor(25, 25, 25))
            pa.setColor(QPalette.ColorRole.AlternateBase, QColor(53, 53, 53))
            pa.setColor(QPalette.ColorRole.ToolTipBase, Qt.GlobalColor.white)
            pa.setColor(QPalette.ColorRole.ToolTipText, Qt.GlobalColor.white)
            pa.setColor(QPalette.ColorRole.Text, Qt.GlobalColor.white)
            pa.setColor(QPalette.ColorRole.Button, QColor(53, 53, 53))
            pa.setColor(QPalette.ColorRole.ButtonText, Qt.GlobalColor.white)
            pa.setColor(QPalette.ColorRole.BrightText, Qt.GlobalColor.red)
            pa.setColor(QPalette.ColorRole.Link, QColor(42, 130, 218))
            pa.setColor(QPalette.ColorRole.Highlight, QColor(42, 130, 218))
            pa.setColor(
                QPalette.ColorRole.HighlightedText, Qt.GlobalColor.black
            )
            qapplication.setStyleSheet(
                "QToolTip { color: #ffffff; background-color: #2a82da; "
                "border: 1px solid white; }"
            )

        else:
            background = QColor("#f2f2f2")
            pa.setColor(QPalette.ColorRole.Window, background)
            pa.setColor(QPalette.ColorRole.AlternateBase, background)
            pa.setColor(QPalette.ColorRole.Button, background)

        qapplication.setPalette(pa)

        try:
            launcher_font_size = int(
                Settings.instance().get("launcher_font_size")
            )
        except ValueError:
            if fusion_variant == "fws":
                launcher_font_size = 13
            else:
                launcher_font_size = 0
        if launcher_font_size:
            print("FONT: Override size {}".format(launcher_font_size))
            font.setPointSize(launcher_font_size)
            if fusion_variant == "fws":
                font = QFont("Roboto")
                font.setPointSizeF(10.5)
                font.setHintingPreference(
                    QFont.HintingPreference.PreferNoHinting
                )
            qapplication.setFont(font)

    import fsui

    if use_fusion_theme:
        fsui.theme = "fusion"
        fsui.theme_variant = fusion_variant



from fsbc.task import Task
from fsgs.context import fsgs
from fsgs.Database import Database
from fsgs.filedatabase import FileDatabase
from fsgs.LockerDatabase import LockerDatabase
from launcher.i18n import gettext

# class MaintenanceSettingsPage(SettingsPage):
#
#     def __init__(self, parent):
#         super().__init__(parent)
#         icon = fsui.Icon("maintenance", "pkg:workspace")
#         title = gettext("Maintenance")
#         subtitle = gettext("Miscellaneous functions to optimize {name}").format(
#                            name="FS-UAE Launcher")
#         self.add_header(icon, title, subtitle)
#
#         label = fsui.MultiLineLabel(self, gettext(
#             "Defragmenting the databases will improve performance "
#             "by ensuring that tables and indices are stored contiguously "
#             "on disk. It will also reclaim some storage space."), 640)
#         self.layout.add(label, fill=True, margin_top=20)
#
#         button = fsui.Button(self, gettext("Defragment Databases"))
#         button.activated.connect(self.on_defragment_button)
#         self.layout.add(button, margin_top=20)
#
#     def on_defragment_button(self):
#         TaskDialog(self.get_window(), DefragmentDatabasesTask()).show()


class DefragmentDatabasesTask(Task):
    def __init__(self):
        Task.__init__(self, gettext("Defragment Databases"))

    def run(self):
        self.defragment(FileDatabase.get_instance(), "Files.sqlite")
        self.stop_check()
        self.defragment(LockerDatabase.instance(), "Locker.sqlite")
        self.stop_check()
        self.defragment(fsgs.get_game_database(), "Amiga.sqlite")
        self.stop_check()
        self.defragment(Database.get_instance(), "Database.sqlite")

    def defragment(self, database, name):
        self.set_progress(gettext("Defragmenting {name}").format(name=name))
        with database:
            cursor = database.cursor()
            cursor.execute("VACUUM")



"""support functions for handling key-value parameters"""

def parse_key_value_string(s, d):
  """parse a single key=value string and add it to the given dictionary"""
  pos = s.find('=')
  if pos == -1:
    d[s] = True
  else:
    key = s[:pos]
    value = s[pos+1:]
    v = value.lower()
    if v in ("true","on"):
      value = True
    elif v in ("false","off"):
      value = False
    elif v.startswith("0x"):
      value = int(v[2:],16)
    else:
      # try a value
      try:
        value = int(value)
      except ValueError:
        pass
    d[key] = value

def parse_key_value_strings(strs):
  """parse an array of strings with key=value contents and create a dictionary from it."""
  result = {}
  for s in strs:
    parse_key_value_string(s, result)
  return result

def parse_keys_values_in_string(s):
  """parse key1=value1,key2=value2,..."""
  return parse_key_value_strings(s.split(','))

def parse_name_args_string(s):
  """parse name:key1=value,... and return name,args_dict"""
  pos = s.find(":")
  if pos == -1:
    return s, {}
  name = s[:pos]
  args = parse_keys_values_in_string(s[pos+1:])
  return name, args



from __future__ import absolute_import
from __future__ import print_function

from ..Block import Block

class RDBPhysicalDrive:
  def __init__(self, cyls=0, heads=0, secs=0, 
               interleave=1, parking_zone=-1, write_pre_comp=-1, reduced_write=-1, step_rate=3):
    if parking_zone == -1:
        parking_zone = cyls
    if write_pre_comp == -1:
        write_pre_comp = cyls
    if reduced_write == -1:
        reduced_write = cyls
    
    self.cyls = cyls
    self.heads = heads
    self.secs = secs
    self.interleave = interleave
    self.parking_zone = parking_zone
    self.write_pre_comp = write_pre_comp
    self.reduced_write = reduced_write
    self.step_rate = step_rate
  
  def dump(self):
    print("PhysicalDrive")
    print(" cyls:           %d" % self.cyls)
    print(" heads:          %d" % self.heads)
    print(" secs:           %d" % self.secs)
    print(" interleave:     %d" % self.interleave)
    print(" parking_zone:   %d" % self.parking_zone)
    print(" write_pre_comp: %d" % self.write_pre_comp)
    print(" reduced_write:  %d" % self.reduced_write)
    print(" step_rate:      %d" % self.step_rate)
  
  def read(self, blk):
    self.cyls = blk._get_long(16)
    self.secs = blk._get_long(17)
    self.heads = blk._get_long(18)
    self.interleave = blk._get_long(19)
    self.parking_zone = blk._get_long(20)
    
    self.write_pre_comp = blk._get_long(24)
    self.reduced_write = blk._get_long(25)
    self.step_rate = blk._get_long(26)
    
  def write(self, blk):
    blk._put_long(16, self.cyls)
    blk._put_long(17, self.secs)
    blk._put_long(18, self.heads)
    blk._put_long(19, self.interleave)
    blk._put_long(20, self.parking_zone)
    
    blk._put_long(24, self.write_pre_comp)
    blk._put_long(25, self.reduced_write)
    blk._put_long(26, self.step_rate)


class RDBLogicalDrive:
  def __init__(self, rdb_blk_lo=0, rdb_blk_hi=0, lo_cyl=0, hi_cyl=0, 
               cyl_blks=0, high_rdsk_blk=0, auto_park_secs=0):
    self.rdb_blk_lo = rdb_blk_lo
    self.rdb_blk_hi = rdb_blk_hi
    self.lo_cyl = lo_cyl
    self.hi_cyl = hi_cyl
    self.cyl_blks = cyl_blks
    self.high_rdsk_blk = high_rdsk_blk
    self.auto_park_secs = auto_park_secs

  def dump(self):
    print("LogicalDrive")
    print(" rdb_blk_lo:     %d" % self.rdb_blk_lo)
    print(" rdb_blk_hi:     %d" % self.rdb_blk_hi)
    print(" lo_cyl:         %d" % self.lo_cyl)
    print(" hi_cyl:         %d" % self.hi_cyl)
    print(" cyl_blks:       %d" % self.cyl_blks)
    print(" high_rdsk_blk:  %d" % self.high_rdsk_blk)
    print(" auto_park_secs: %d" % self.auto_park_secs)

  def read(self, blk):
    self.rdb_blk_lo = blk._get_long(32)
    self.rdb_blk_hi = blk._get_long(33)
    self.lo_cyl = blk._get_long(34)
    self.hi_cyl = blk._get_long(35)
    self.cyl_blks = blk._get_long(36)
    self.auto_park_secs = blk._get_long(37)
    self.high_rdsk_blk = blk._get_long(38)

  def write(self, blk):
    blk._put_long(32, self.rdb_blk_lo)
    blk._put_long(33, self.rdb_blk_hi)
    blk._put_long(34, self.lo_cyl)
    blk._put_long(35, self.hi_cyl)
    blk._put_long(36, self.cyl_blks)
    blk._put_long(37, self.auto_park_secs)
    blk._put_long(38, self.high_rdsk_blk)


class RDBDriveID:
  def __init__(self, disk_vendor="", disk_product="", disk_revision="", 
               ctrl_vendor="", ctrl_product="", ctrl_revision=""):
    self.disk_vendor = disk_vendor
    self.disk_product = disk_product
    self.disk_revision = disk_revision
    self.ctrl_vendor = ctrl_vendor
    self.ctrl_product = ctrl_product
    self.ctrl_revision = ctrl_revision
    
  def dump(self):
    print("DriveID")
    print(" disk_vendor:    '%s'" % self.disk_vendor)
    print(" disk_product:   '%s'" % self.disk_product)
    print(" disk_revision:  '%s'" % self.disk_revision)
    print(" ctrl_vendor:    '%s'" % self.ctrl_vendor)
    print(" ctrl_product:   '%s'" % self.ctrl_product)
    print(" ctrl_revision:  '%s'" % self.ctrl_revision)
    
  def read(self, blk):
    self.disk_vendor = blk._get_cstr(40, 8)
    self.disk_product = blk._get_cstr(42, 16)
    self.disk_revision = blk._get_cstr(46, 4)
    self.ctrl_vendor = blk._get_cstr(47, 8)
    self.ctrl_product = blk._get_cstr(49, 16)
    self.ctrl_revision = blk._get_cstr(53, 4)

  def write(self, blk):
    blk._put_cstr(40, 8, self.disk_vendor)
    blk._put_cstr(42, 16, self.disk_product)
    blk._put_cstr(46, 4, self.disk_revision)
    blk._put_cstr(47, 8, self.ctrl_vendor)
    blk._put_cstr(49, 16, self.ctrl_product)
    blk._put_cstr(53, 4, self.ctrl_revision)


class RDBlock(Block):
  def __init__(self, blkdev, blk_num=0):
    Block.__init__(self, blkdev, blk_num, chk_loc=2, is_type=Block.RDSK)
  
  def create(self, phy_drv, log_drv, drv_id,
             host_id=7, block_size=512, flags=0x17, 
             badblk_list=Block.no_blk, part_list=Block.no_blk, fs_list=Block.no_blk, init_code=Block.no_blk,
             size=64):
    Block.create(self)
    self.size = size
    self.host_id = host_id
    self.block_size = block_size
    self.flags = flags
    
    self.badblk_list = badblk_list
    self.part_list = part_list
    self.fs_list = fs_list
    self.init_code = init_code

    self.phy_drv = phy_drv
    self.log_drv = log_drv
    self.drv_id = drv_id
    self.valid = True
      
  def write(self):
    self._create_data()
    
    self._put_long(1, self.size)
    self._put_long(3, self.host_id)
    self._put_long(4, self.block_size)
    self._put_long(5, self.flags)
    
    self._put_long(6, self.badblk_list)
    self._put_long(7, self.part_list)
    self._put_long(8, self.fs_list)
    self._put_long(9, self.init_code)
    
    self.phy_drv.write(self)
    self.log_drv.write(self)
    self.drv_id.write(self)
    
    Block.write(self)
  
  def read(self):
    Block.read(self)
    if not self.valid:
      return False
    
    self.size = self._get_long(1)
    self.host_id = self._get_long(3)
    self.block_size = self._get_long(4)
    self.flags = self._get_long(5)
    
    self.badblk_list = self._get_long(6)
    self.part_list = self._get_long(7)
    self.fs_list = self._get_long(8)
    self.init_code = self._get_long(9)
    
    self.phy_drv = RDBPhysicalDrive()
    self.phy_drv.read(self)
    self.log_drv = RDBLogicalDrive()
    self.log_drv.read(self)
    self.drv_id = RDBDriveID()
    self.drv_id.read(self)
    
    return self.valid
  
  def dump(self):
    Block.dump(self, "RigidDisk")
    
    print(" size:           %d" % self.size)
    print(" host_id:        %d" % self.host_id)
    print(" block_size:     %d" % self.block_size)
    print(" flags:          0x%08x" % self.flags)
    print(" badblk_list:    %s" % self._dump_ptr(self.badblk_list))
    print(" part_list:      %s" % self._dump_ptr(self.part_list))
    print(" fs_list:        %s" % self._dump_ptr(self.fs_list))
    print(" init_code:      %s" % self._dump_ptr(self.init_code))
    
    self.phy_drv.dump()
    self.log_drv.dump()
    self.drv_id.dump()



import weakref

from arcade.glui.animation import AnimationSystem
from arcade.glui.input import InputHandler
from arcade.glui.state import State
from arcade.glui.texturemanager import TextureManager


class Callbacks:
    def __init__(self):
        self.width = 320
        self.height = 200
        self.window = None
        self.idle_counter = 0

    def initialize(self, width, height):
        self.width = width
        self.height = height

        # Re-create state object
        State.reset()

        TextureManager.reset()

        import arcade.glui.window

        arcade.glui.window.main_window = self
        arcade.glui.window.show()

    def active(self):
        """Check whether the user interface is active or not. If not active,
        we do not want to force regular/timed refresh event."""
        state = State.get()
        # print("state.game_running", state.game_running)
        if state.game_running:
            # print("inactive")
            return False
        # print(state.idle_from)
        # if state.idle_from and state.idle_from < get_current_time():
        # print(InputHandler.peek_button())
        if not InputHandler.peek_button():
            if not state.dirty:
                if not AnimationSystem.is_active():
                    self.idle_counter += 1
                    if self.idle_counter < 16:
                        # We render 1 frame per second when "idling". We need
                        # to "render" regularly, because some state is updated
                        # during the render phase.
                        return False
        self.idle_counter = 0
        return True

    def resize(self, width, height):
        print("Callbacks.resize", width, height)
        from arcade.glui.window import on_resize

        self.width = width
        self.height = height
        on_resize((width, height))

    def render(self):
        # print("render")
        from arcade.glui.window import main_loop_iteration

        if main_loop_iteration():
            print("main_loop_iteration signalled quit")
            TextureManager.get().unload_textures()
            self.window().quit()

    def timer(self):
        InputHandler.update()

    def restore_window_if_necessary(self):
        pass
        # noinspection PyCallingNonCallable
        # window = self.window()
        # under Gnome 3, the window is minized when launching FS-UAE
        # full-screen from full-screen arcade/game center.
        # window.restore_window_if_necessary()

    def set_window(self, window):
        self.window = weakref.ref(window)



class RequestParams:
    def __init__(self):
        self.platform = ""
        self.game_uuid = ""
        self.state = ""
        self.fullscreen = ""
        self.fullscreen_mode = ""
        self.window_border = ""



from oyoyo import helpers
from oyoyo.cmdhandler import DefaultCommandHandler


class CommandHandler(DefaultCommandHandler):
    # noinspection SpellCheckingInspection
    def nicknameinuse(self, nick, *message):
        print("Nickname in use", nick, message)
        helpers.nick(self.client, self.client.handler.generate_nick())

    def __unhandled__(self, command, *args):
        handler = self.client.handler
        handler.post_message(command, args)



import os

from fsgs.FSGSDirectories import FSGSDirectories


class PluginHelper(object):
    def __init__(self):
        pass

    def find_resource_dirs(self, resource_name):
        resource_dirs = []
        for plugin in self.find_plugins_with_resource(resource_name):
            resource_dir = os.path.join(plugin, resource_name)
            resource_dirs.append(resource_dir)
        return resource_dirs

    @staticmethod
    def find_plugins():
        plugins = []
        plugins_dir = FSGSDirectories.get_plugins_dir()
        if plugins_dir is None:
            return plugins
        for item in os.listdir(plugins_dir):
            plugin_dir = os.path.join(plugins_dir, item)
            if not os.path.isdir(plugin_dir):
                continue
            if os.path.exists(os.path.join(plugin_dir, "disabled")):
                continue
            plugins.append(plugin_dir)
        return plugins

    def find_plugins_with_resource(self, resource_name):
        plugins = []
        for plugin_dir in self.find_plugins():
            resource_dir = os.path.join(plugin_dir, resource_name)
            if os.path.exists(resource_dir):
                plugins.append(plugin_dir)
        return plugins



import threading
import traceback

import time

import workspace.sys
from workspace import ui


COLOR = (0x44, 0x44, 0x44)


class ShellWindow(ui.Window):
    def __init__(self, parent, argv=None):
        super().__init__(
            parent, "Shell", menu=True, maximizable=False, color=COLOR
        )
        if argv is None or len(argv) == 0:
            argv = ["CLI"]

        col = ui.Column()
        self.add(col)
        # col.add(ui.TitleSeparator(self))
        self.canvas = ShellWidget(self, argv=argv)
        self.set_background_color(ui.Color(*COLOR))
        col.add(self.canvas, left=12, top=6, right=12, bottom=6)

        self.closed.connect(self.__on_closed)

    def __on_closed(self):
        self.canvas.terminate_process()


class ShellWidget(ui.Canvas):
    output = ui.Signal()
    exited = ui.Signal()

    def __init__(self, parent, *, argv, columns=80, rows=24):
        super().__init__(parent)
        self.font = ui.Font("Roboto Mono", 14)
        self.set_background_color(ui.Color(*COLOR))
        self.text_color = ui.Color(0xFF, 0xFF, 0xFF)

        self.input_thread = None
        self.output_thread = None
        self.output.connect(self.__on_output)

        self._completed = False
        self.exited.connect(self.__on_exited)

        self.line_buffer = []
        self.lines = []
        # for _ in range(25):
        self._new_line()
        self.cx = 0
        self.cy = 0

        # self.execute(["version"])
        # self.execute(["python", "-i"])
        self.process = None
        self.execute(argv)

        # FIXME: Using Qt directly
        from fsui.qt import Qt, QPainter, QPixmap

        self._widget.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

        painter = QPainter()
        pixmap = QPixmap(100, 100)
        painter.begin(pixmap)
        painter.setFont(self.font.qfont())
        size = painter.boundingRect(
            0, 0, 10000, 1000, Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignTop, " " * columns
        )
        painter.end()
        print(size)
        self.line_width = size.width()
        self.line_height = size.height()

        self.set_min_size((self.line_width, self.line_height * rows))

    def execute(self, args):
        self.process = workspace.sys.workspace_exec(args)
        print(self.process)
        self.input_thread = InputThread(self, self.process)
        self.input_thread.start()
        self.output_thread = OutputThread(self, self.process)
        self.output_thread.start()

    def terminate_process(self):
        self.input_thread.stop()
        self.output_thread.stop()
        self.process.kill()

    def _new_line(self):
        self.lines = self.lines[-24:]
        self.lines.append([" "] * 80)
        self.cx = 0
        self.cy = len(self.lines) - 1

    def __on_output(self):
        print("on_output")
        # This runs in the main thread
        data = self.output_thread.data()
        print(data)
        for byte in data:
            self._print_byte(byte)

    def __on_exited(self):
        self._complete()

    def _complete(self):
        if not self._completed:
            self._print_text("\n[Process completed]")
            self.input_thread.stop()
            self.output_thread.stop()
            self._completed = True

    def _print_text(self, text):
        for char in text:
            self._print_char(char)

    def _print_byte(self, byte):
        # FIXME: UTF-8...
        char = byte.decode("ISO-8859-1")
        self._print_char(char)

    def _print_char(self, char):
        if char == "\n":
            self._new_line()
            return
        if char == "\r":
            self.cx = 0
            return
        # print(self.cy, self.cx)
        self.lines[self.cy][self.cx] = char
        # line = self.lines[-1]
        self.cx += 1
        if self.cx == 80:
            # self.cy += 1
            # self.cx = 0

            # if self.cy == 25:
            #    self.lines = self.lines[:24]

            self._new_line()

        # if len(line) < 80:
        #     self.lines[-1] = line + char
        # else:
        #      line = char
        #      self.lines.append(line)
        #      if len(self.lines) > 25:
        #         self.lines = self.lines[-25:]
        self.refresh()

    def _erase_char(self):
        # FIXME: TODO: multi-line erase..
        if self.cx > 0:
            self.cx -= 1
            self.lines[self.cy][self.cx] = " "
        self.refresh()

    def on_paint(self):
        painter = ui.Painter(self)
        painter.set_font(self.font)
        if hasattr(self, "text_color"):
            painter.set_text_color(self.text_color)
        # lines = self.buffer()
        x, y = 0, 0
        for line_data in self.lines:
            line = "".join(line_data)
            # tw, th = painter.measure_text(line)
            painter.draw_text(line, x, y)
            # y += th
            y += self.line_height

    def on_key_press(self, event):
        print("on_key_press", event)
        if self._completed:
            print("process is completed, ignoring")
            return
        # FIXME: using a QKeyEvent directly
        from fsui.qt import QKeyEvent

        assert isinstance(event, QKeyEvent)
        char = event.text()
        print(repr(char))
        if not char:
            print("char was", repr(char))
            return
        if char == "\x08":
            if len(self.line_buffer) > 0:
                self.line_buffer = self.line_buffer[:-1]
                self._erase_char()
            return
        if char == "\x04":
            self.input_thread.add_byte(b"\x04")
            # FIXME: Close stdin if/when you type Ctrl+D?
            # self._p.stdin.close()
            return
        if char == "\r":
            # Qt peculiarity...
            char = "\n"
            self.line_buffer.append(char)
            self._new_line()
            bytes = []
            for char in self.line_buffer:
                bytes.append(char.encode("ISO-8859-1"))
            self.input_thread.add_bytes(bytes)
            self.line_buffer.clear()
            return
        else:
            self.line_buffer.append(char)
            self._print_char(char)
        # byte = char.encode("ISO-8859-1")
        # self.input_thread.add_byte(byte)


class OutputThread(threading.Thread):
    def __init__(self, widget, p):
        super().__init__()
        self._widget = widget
        self._p = p
        self._data = []
        self._available = False
        self._lock = threading.Lock()
        self.stop_flag = False

    def data(self):
        with self._lock:
            data = self._data
            self._data = []
            self._available = False
            return data

    def run(self):
        print("OutputThread.run")
        try:
            self._run()
        except Exception:
            traceback.print_exc()
        self._widget = None

    def stop(self):
        self.stop_flag = True
        self._widget = None

    def _run(self):
        while True:
            byte = self._p.stdout.read(1)
            if not byte:
                print("no more data, ending output thread")
                if self._widget is not None:
                    self._widget.exited.emit()
                return
            with self._lock:
                self._data.append(byte)
                if not self._available:
                    self._available = True
                    print("emit time")
                    if self._widget is not None:
                        self._widget.output.emit()


class InputThread(threading.Thread):
    """Handles data sent to child process"""

    def __init__(self, widget, p):
        super().__init__()
        self._widget = widget
        self._p = p
        self._data = []
        # self._available = False
        self._lock = threading.Lock()
        self.stop_flag = False

    def add_byte(self, byte):
        with self._lock:
            self._data.append(byte)

    def add_bytes(self, bytes):
        with self._lock:
            self._data.extend(bytes)

    def run(self):
        print("InputThread.run")
        try:
            self._run()
        except Exception:
            traceback.print_exc()
        self._widget = None

    def stop(self):
        self.stop_flag = True
        self._widget = None

    def _run(self):
        while not self.stop_flag:
            # FIXME: conditions instead
            time.sleep(0.01)
            with self._lock:
                if len(self._data) == 0:
                    continue
                data = b"".join(self._data)
                self._data = []
            print("writing", data)
            self._p.stdin.write(data)
            self._p.stdin.flush()



import logging
import os

import fsboot

logger = logging.getLogger("WORKSPACE")
assigns = {
    "C": "Workspace:C",
    # "CLIPS": "Ram Disk:Clipboards",
    "DEVS": "Workspace:Devs",
    "ENV": "Ram Disk:ENV",
    "ENVARC": "Workspace:Prefs/Env-Archive",
    "FONTS": "Workspace:Fonts",
    "HELP": "Workspace:Locale/Help",
    # "KEYMAPS": "Workspace:Devs/Keymaps",
    # "L": "Workspace:L",
    "LIBS": "Workspace:Libs",
    "LOCALE": "Workspace:Locale",
    # "PRINTERS": "Workspace:Devs/Printers",
    "REXX": "Workspace:S",
    "S": "Workspace:S",
    "SYS": "Workspace:",
    # "T": "Ram Disk:T",
    # FIXME: Perhaps
    "CDROMS": "Media/CD-ROMs",
    "DISKS": "Media/Hard Drives",
    "FLOPPIES": "Media/Floppies",
    "ICONS": "Workspace:FIXME/Icons",
    "LOGS": "Ram Disk:T/Logs",
    # "DATABASES": "Data/Databases",
    # Special
    "VOLUMES": ".",
}


def expand_assigns(path):
    logger.debug("expand_assigns %s", repr(path))
    volume, p = path.split(":")
    try:
        dir = assigns[volume]
        if not dir.endswith(":"):
            dir += "/"
        return dir + p
    except KeyError:
        return path


def basename(path):
    if "/" in path:
        return path.rsplit("/", 1)[1]
    if ":" in path:
        return path.rsplit(":", 1)[1]
    return path


def extension(path, prefix=False, suffix=True):
    name = basename(path)
    ext = ""
    name = name.lower()
    if suffix:
        ext_2 = os.path.splitext(name)[1]
        if ext_2:
            assert ext_2[0] == "."
            ext = ext_2[1:]
    if prefix:
        if name.startswith("mod."):
            ext = "mod"
    return ext


def exists(path):
    return os.path.exists(host(path))


def host(path):
    path = expand_assigns(path)
    # FIXME: make absolute
    path = path.replace("AmigaForever:", "AmigaForever/")
    path = path.replace("AmigaOS4.1:", "AmigaOS4.1/")
    path = path.replace("AmiKit:", "AmiKit/")
    path = path.replace("ClassicWB:", "ClassicWB/")
    path = path.replace("Data:", "Data/")
    path = path.replace("Games:", "Games/")
    path = path.replace("Media:", "Media/")
    path = path.replace("Ram Disk:", "Ram Disk/")
    path = path.replace("Shared:", "Shared/")
    path = path.replace("Systems:", "Systems/")
    path = path.replace("Workspace:", "Workspace/")
    if not os.path.isabs(path):
        path = os.path.join(fsboot.base_dir(), path)
    return path


def listdir(path):
    host_path = host(path)
    items = os.listdir(host_path)
    if path == "VOLUMES:":
        items = [x for x in items if os.path.isdir(os.path.join(host_path, x))]
        if "Cache" in items:
            items.remove("Cache")
        if "Floppies" in items:
            items.remove("Floppies")
        if "CD-ROMs" in items:
            items.remove("CD-ROMs")
        if "Hard Drives" in items:
            items.remove("Hard Drives")
        if "Kickstarts" in items:
            items.remove("Kickstarts")
        if "Save States" in items:
            items.remove("Save States")
        if "Controllers" in items:
            items.remove("Controllers")
        if "Source" in items:
            items.remove("Source")
        if "Themes" in items:
            items.remove("Themes")
        if "Configurations" in items:
            items.remove("Configurations")
    return items


def isdir(path):
    logger.debug("workspace.path.isdir %s", repr(path))
    return os.path.isdir(host(path))


def join(a, b, *args):
    if a == "VOLUMES:":
        return b + ":"
    path = a
    if not a.endswith(":"):
        path += "/"
    path += b
    for arg in args:
        path += "/" + arg
    return path


class Path:
    def __init__(self, path):
        self._path = path

    def host(self):
        return host(self._path)



import threading
import time
import traceback
from uuid import uuid4

from fsgs.input.enumeratehelper import EnumerateHelper
from fsgs.platform import PlatformHandler


class Instance:
    STATE_INITIALIZED = "INITIALIZED"
    STATE_PREPARING = "PREPARING"
    STATE_READY = "READY"
    STATE_RUNNING = "RUNNING"
    STATE_STOPPED = "STOPPED"
    STATE_FINALIZING = "FINALIZING"
    STATE_FINALIZED = "FINALIZED"

    def __init__(self, fsgs, instance_uuid):
        self.fsgs = fsgs
        self.uuid = instance_uuid
        self.thread = None
        self.runner = None
        self._state = self.STATE_INITIALIZED

    def get_state(self):
        return self._state

    def set_state(self, state):
        self._state = state

    state = property(get_state, set_state)

    def start(self):
        assert self.thread is None

        platform_handler = PlatformHandler.create(self.fsgs.game.platform.id)
        self.runner = platform_handler.get_runner(self.fsgs)

        device_helper = EnumerateHelper()
        device_helper.default_port_selection(self.runner.ports)

        self.thread = threading.Thread(target=self._thread)
        self.thread.start()

    def _thread(self):
        try:
            while self.state == self.STATE_INITIALIZED:
                time.sleep(0.1)
            assert self.state == self.STATE_PREPARING

            self.runner.prepare()
            self.state = self.STATE_READY

            while self.state == self.STATE_READY:
                time.sleep(0.1)
            assert self.state == self.STATE_RUNNING

            process = self.runner.run()
            process.wait()
            self.state = self.STATE_STOPPED

            while self.state == self.STATE_STOPPED:
                time.sleep(0.1)
            assert self.state == self.STATE_FINALIZING

            self.runner.finish()
            self.state = self.STATE_FINALIZED
        except Exception:
            print("EXCEPTION IN Instance.thread")
            traceback.print_exc()

    def destroy(self):
        print("Instance.destroy")


class Session:
    def __init__(self, fsgs):
        self.fsgs = fsgs

    def create_instance(self):
        assert self.fsgs.game.uuid
        assert self.fsgs.game.variant.uuid

        instance = Instance(self.fsgs, str(uuid4()))
        return instance



from fsgs.amiga.amiga import Amiga
from fsgs.option import Option

CONFIG_KEY_BLACKLIST = [
    "_*",
    "amigamemo_url",
    "config_feature",
    "config_name",
    "config_search",
    "configurations_dir_mtime",
    "database_arcade",
    "database_dos",
    "database_gba",
    "database_locker",
    "database_nes",
    "database_show_adult",
    "database_show_games",
    "database_snes",
    "database_url",
    "developer",
    "front_sha1",
    "hol_url",
    "irc_nick",
    "kickstarts_dir_mtime",
    "languages",
    "last_cd_dir",
    "last_floppy_dir",
    "last_hd_dir",
    "last_rom_dir",
    "last_scan",
    "last_settings_page",
    Option.LAUNCHER_THEME,
    "lemon_url",
    "mobygames_url",
    "netplay_feature",
    "parent_uuid",
    "platform",
    "players",
    "primary_joystick",
    "publisher",
    "screen1_sha1",
    "screen2_sha1",
    "screen3_sha1",
    "screen4_sha1",
    "screen5_sha1",
    "search_path",
    "title_sha1",
    "variant_rating",
    "variant_uuid",
    "whdload_url",
    "wikipedia_url",
    "x_*",
    "year",
]


class ConfigWriter(object):
    def __init__(self, config):
        self.config = config

    def create_fsuae_config(self):
        print("create_fsuae_config")
        c = []

        num_drives = 0
        for i in range(Amiga.MAX_FLOPPY_DRIVES):
            key = "floppy_drive_{0}".format(i)
            value = self.config.get(key)
            if value:
                num_drives = i + 1
        num_drives = max(1, num_drives)

        print("")
        print("-------------" * 6)
        print("CONFIG")
        c.append("[fs-uae]")
        for key in sorted(self.config.keys()):
            ignore = False
            if key.startswith("floppy_drive_"):
                for i in range(4):
                    if key == "floppy_drive_{0}".format(i):
                        if i >= num_drives:
                            ignore = True
                            break
            normalized_key = key.lower().replace("-", "_")
            for ignore_key in CONFIG_KEY_BLACKLIST:
                if ignore_key == normalized_key or (
                    ignore_key.endswith("*")
                    and normalized_key.startswith(ignore_key[:-1])
                ):
                    ignore = True
            if ignore:
                continue
            value = self.config[key]
            print(key, repr(value))
            value = value.replace("\\", "\\\\")
            if value:
                c.append("{0} = {1}".format(key, value))

        return c



import traceback

import fsui
from launcher.ui.config.model import create_model


class ConfigBrowser(fsui.VerticalItemView):
    def __init__(self, parent):
        # fsui.VerticalItemView.__init__(self, parent, border=(not Skin.fws()))
        fsui.VerticalItemView.__init__(self, parent, border=False)
        self.items = []
        self.game_icon = fsui.Image("launcher:res/16x16/controller.png")
        self.config_icon = fsui.Image("launcher:res/fsuae_config_16.png")

        self.manual_download_icon = fsui.Image(
            "launcher:res/16x16/arrow_down_yellow.png"
        )
        self.auto_download_icon = fsui.Image(
            "launcher:res/16x16/arrow_down_green.png"
        )
        self.blank_icon = fsui.Image("launcher:res/16x16/blank.png")
        self.missing_color = fsui.Color(0xA8, 0xA8, 0xA8)
        self.platform_icons = {}

        from workspace.ui.theme import WorkspaceTheme

        theme = WorkspaceTheme.instance()
        self.set_row_height(28)
        # self.list_view.set_background_color(fsui.Color(0xeb, 0xeb, 0xeb))
        self.setStyleSheet(
            """
        QListView {{
            background-color: {0};
            outline: none;
        }}
        QListView::item {{
            padding-left: 6px;
            border: 0px;
        }}
        QListView::item:selected {{
            background-color: {1};
        }}
        """.format(
                theme.sidebar_background.to_hex(),
                theme.selection_background.to_hex(),
            )
        )

    def update_from_implicit(self, implicit):
        def flatten(item_list, level):
            for model_item in item_list:
                if model_item.active:
                    # text = "    " * level + str(model_item.text)
                    text = str(model_item.text)
                    if model_item.extra:
                        text += " [" + model_item.extra + "]"
                    items.append(text)
                if model_item.children:
                    flatten(model_item.children, level + 1)

        items = []
        try:
            model = create_model(implicit, show_all=False)
        except Exception:
            traceback.print_exc()
            print("create_model failed")
        else:
            flatten(model.items, 0)
        self.set_items(items)

    def on_select_item(self, index):
        if index is None:
            return
        pass

    def on_activate_item(self, index):
        pass

    def set_items(self, items):
        self.items = items
        self.update()

    def get_item_count(self):
        return len(self.items)

    def get_item_text(self, index):
        item = self.items[index]
        return item
        # return ""

    # def get_item_search_text(self, index):
    #     # return self.items[index][3]
    #     # FIXME: lower-case search string?
    #     return self.items[index][str("sort_key")]
    #
    # def get_item_text_color(self, index):
    #     have = self.items[index][str("have")]
    #     if not have:
    #         return self.missing_color

    def get_item_icon(self, index):
        # item = self.items[index]
        # platform_id = (item[str("platform")] or "").lower()
        # if item[str("have")] == 1:
        #     return self.manual_download_icon
        # elif item[str("have")] == 0:
        #     return self.blank_icon
        # elif item[str("have")] == 2:
        #     return self.auto_download_icon
        # elif item[str("have")] == 4:
        #     if platform_id not in self.platform_icons:
        #         try:
        #             icon = fsui.Image("launcher:res/16x16/{0}.png".format(
        #                 platform_id))
        #         except Exception:
        #             icon = self.game_icon
        #         self.platform_icons[platform_id] = icon
        #     return self.platform_icons[platform_id]
        # else:
        return self.config_icon

    # def on_get_item_tooltip(self, row, column):
    #     return ""



import weakref

from fsbc.signal import Signal

from .contextaware import ContextAware


class SignalBehavior:
    def __init__(self, context, parent, names):
        parent.__signal_enable_behavior = self
        self._context = context
        self._parent = weakref.ref(parent)
        self._names = set(names)
        try:
            parent.destroyed.connect(self.on_parent_destroyed)
        except AttributeError:
            print(
                "WARNING:SignalBehavior without remove_listener implementation"
            )
        for signal in self._names:
            self._context.connect(
                signal, getattr(self._parent(), "on_{}_signal".format(signal))
            )
            # self._context.connect(signal, self._parent())

    def on_parent_destroyed(self):
        for signal in self._names:
            self._context.disconnect(
                signal, getattr(self._parent(), "on_{}_signal".format(signal))
            )
            # self._context.disconnect(signal, self._parent())


# noinspection PyMethodMayBeStatic
class SignalContext(ContextAware):
    def __init__(self, context):
        ContextAware.__init__(self, context)

    def signal_name(self, signal):
        # FIXME: use fsgs-context-instance-specific signals
        # return "fsgs:{}{}".format(id(signal), signal)
        return "fsgs:{}".format(signal)

    def connect(self, signal, listener):
        Signal(self.signal_name(signal)).connect(listener)

    def disconnect(self, signal, listener):
        Signal(self.signal_name(signal)).disconnect(listener)

    def emit(self, signal, args):
        Signal(self.signal_name(signal)).notify(*args)

    # FIXME: Deprecated
    def notify(self, signal, args):
        return self.emit(signal, args)

    def process(self):
        Signal.process_all_signals()

    def add_behavior(self, parent, names):
        SignalBehavior(self, parent, names)



import fsui as fsui
from .WidgetTestCase import WidgetTestCase


class TestTextField(WidgetTestCase):
    def test_text_field(self):
        self.setup_dialog()
        text_field = fsui.TextField(self.dialog)
        text_field.set_text("Hello World")
        text_field.disable()
        text_field.enable()
        self.run_dialog()

    def test_text_field_with_text(self):
        self.setup_dialog()
        text_field = fsui.TextField(self.dialog, "Hello World")
        self.assertEqual(text_field.get_text(), "Hello World")
        self.run_dialog()



# Copyright (c) 2008 Duncan Fordyce
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import socket
import time
import traceback

from oyoyo.parse import *
from oyoyo import helpers
from oyoyo.cmdhandler import CommandError


class IRCClientError(Exception):
    pass


class IRCClient:
    """ IRC Client class. This handles one connection to a server.
    This can be used either with or without IRCApp ( see connect() docs )
    """

    def __init__(self, cmd_handler, **kwargs):
        """ the first argument should be an object with attributes/methods named
        as the irc commands. You may subclass from one of the classes in
        oyoyo.cmdhandler for convenience but it is not required. The
        methods should have arguments (prefix, args). prefix is
        normally the sender of the command. args is a list of arguments.
        Its recommened you subclass oyoyo.cmdhandler.DefaultCommandHandler,
        this class provides defaults for callbacks that are required for
        normal IRC operation.

        all other arguments should be keyword arguments. The most commonly
        used will be nick, host and port. You can also specify an "on connect"
        callback. ( check the source for others )

        Warning: By default this class will not block on socket operations, this
        means if you use a plain while loop your app will consume 100% cpu.
        To enable blocking pass blocking=True.

        >>> from oyoyo import helpers        >>> class My_Handler(DefaultCommandHandler):
        ...     def privmsg(self, prefix, command, args):
        ...         print "%s said %s" % (prefix, args[1])
        ...
        >>> def connect_callback(c):
        ...     helpers.join(c, '#myroom')
        ...
        >>> cli = IRCClient(My_Handler,
        ...                 host="irc.freenode.net",
        ...                 port=6667,
        ...                 nick="myname",
        ...                 connect_cb=connect_callback)
        ...
        >>> cli_con = cli.connect()
        >>> while 1:
        ...     cli_con.next()
        ...
        """
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.nick = None
        self.real_name = None
        self.host = None
        self.port = None
        self.connect_cb = None
        self.blocking = False

        self.__dict__.update(kwargs)
        self.command_handler = cmd_handler(self)

        self._end = 0

    def send(self, *args, **kwargs):
        """ send a message to the connected server. all arguments are joined
        with a space for convenience, for example the following are identical

        >>> cli.send("JOIN %s" % some_room)
        >>> cli.send("JOIN", some_room)

        In python 2, all args must be of type str or unicode, *BUT* if they are
          unicode they will be converted to str with the encoding specified by
          the 'encoding' keyword argument (default 'utf8').
        In python 3, all args must be of type str or bytes, *BUT* if they are
          str they will be converted to bytes with the encoding specified by the
          'encoding' keyword argument (default 'utf8').
        """
        # Convert all args to bytes if not already
        encoding = kwargs.get('encoding') or 'utf8'
        bargs = []
        for arg in args:
            if isinstance(arg, str):
                bargs.append(bytes(arg, encoding))
            elif isinstance(arg, bytes):
                bargs.append(arg)
            elif type(arg).__name__ == 'unicode':
                bargs.append(arg.encode(encoding))
            else:
                raise IRCClientError('Refusing to send one of the args from provided: %s'
                                     % repr([(type(arg), arg) for arg in args]))

        msg = bytes(" ", "ascii").join(bargs)
        logging.info('---> send "%s"' % msg)
        self.socket.send(msg + bytes("\r\n", "ascii"))

    def connect(self):
        """ initiates the connection to the server set in self.host:self.port
        and returns a generator object.

        >>> cli = IRCClient(my_handler, host="irc.freenode.net", port=6667)
        >>> g = cli.connect()
        >>> while 1:
        ...     g.next()

        """
        try:
            logging.info('connecting to %s:%s' % (self.host, self.port))
            self.socket.connect(("%s" % self.host, self.port))
            if self.blocking:
                # this also overrides default timeout
                self.socket.setblocking(1)
            else:
                self.socket.setblocking(0)

            helpers.nick(self, self.nick)
            helpers.user(self, self.nick, self.real_name)

            if self.connect_cb:
                self.connect_cb(self)

            buffer = bytes()
            while not self._end:
                try:
                    buffer += self.socket.recv(1024)
                except socket.error as e:
                    try:  # a little dance of compatibility to get the errno
                        errno = e.errno
                    except AttributeError:
                        errno = e[0]
                    if not self.blocking and errno == 11:
                        pass
                    else:
                        raise e
                else:
                    data = buffer.split(bytes("\n", "ascii"))
                    buffer = data.pop()

                    for el in data:
                        prefix, command, args = parse_raw_irc_command(el)
                        try:
                            self.command_handler.run(command, prefix, *args)
                        except CommandError:
                            # error will of already been loggingged by the handler
                            pass

                yield True
        finally:
            if self.socket:
                logging.info('closing socket')
                self.socket.close()


# noinspection PyPep8Naming
class IRCApp:
    """ This class manages several IRCClient instances without the use of threads.
    (Non-threaded) Timer functionality is also included.
    """

    class _ClientDesc:
        def __init__(self, **kwargs):
            self.con = None
            self.autoreconnect = False
            self.__dict__.update(kwargs)

    def __init__(self):
        self._clients = {}
        self._timers = []
        self.running = False
        self.sleep_time = 0.5

    def addClient(self, client, autoreconnect=False):
        """ add a client object to the application. setting autoreconnect
        to true will mean the application will attempt to reconnect the client
        after every disconnect. you can also set autoreconnect to a number
        to specify how many reconnects should happen.

        warning: if you add a client that has blocking set to true,
        timers will no longer function properly """
        logging.info('added client %s (ar=%s)' % (client, autoreconnect))
        self._clients[client] = self._ClientDesc(autoreconnect=autoreconnect)

    def addTimer(self, seconds, cb):
        """ add a timed callback. accuracy is not specified, you can only
        garuntee the callback will be called after seconds has passed.
        ( the only advantage to these timers is they dont use threads )
        """
        assert callable(cb)
        logging.info('added timer to call %s in %ss' % (cb, seconds))
        self._timers.append((time.time() + seconds, cb))

    def run(self):
        """ run the application. this will block until stop() is called """
        # TODO: convert this to use generators too?
        self.running = True
        while self.running:
            found_one_alive = False

            for client, clientdesc in self._clients.items():
                if clientdesc.con is None:
                    clientdesc.con = client.connect()

                try:
                    clientdesc.con.next()
                except Exception as e:
                    logging.error('client error %s' % e)
                    logging.error(traceback.format_exc())
                    if clientdesc.autoreconnect:
                        clientdesc.con = None
                        if isinstance(clientdesc.autoreconnect, (int, float)):
                            clientdesc.autoreconnect -= 1
                        found_one_alive = True
                    else:
                        clientdesc.con = False
                else:
                    found_one_alive = True

            if not found_one_alive:
                logging.info('nothing left alive... quiting')
                self.stop()

            now = time.time()
            timers = self._timers[:]
            self._timers = []
            for target_time, cb in timers:
                if now > target_time:
                    logging.info('calling timer cb %s' % cb)
                    cb()
                else:
                    self._timers.append((target_time, cb))

            time.sleep(self.sleep_time)

    def stop(self):
        """ stop the application """
        self.running = False







import threading

import bottle


class Request(object):
    def __init__(self):
        pass


def get_request():
    return _thread_local.request


class Proxy(object):
    def __init__(self, func):
        self._func = func

    def __getattr__(self, name):
        return getattr(self._func(), name)

    def __setattr__(self, name, value):
        return setattr(self._func(), name, value)


class WebApp(object):
    def __init__(self):
        pass

    def route(self, path):
        def wrapper(func):
            def wrapper2():
                print("wrapper", func)
                return func()

            return bottle.route(path)(wrapper2)

        return wrapper


_thread_local = threading.local()
# request = Proxy(get_request)



import fsui
from launcher.ui.skin import LauncherTheme
from launcher.ui.widgets import CloseButton

from ...i18n import gettext
from ...launcher_signal import LauncherSignal
from .ImportTask import ImportTask

TIMER_INTERVAL = 100


class ImportDialog(fsui.Window):
    AMIGA_FOREVER = 1

    def __init__(self, parent, path, import_type):
        if import_type == self.AMIGA_FOREVER:
            title = gettext("Import From Amiga Forever CD/DVD")
        else:
            title = gettext("Import Kickstarts and ROMs")
        super().__init__(parent, title, maximizable=False)
        self.theme = LauncherTheme.get()
        self.layout = fsui.VerticalLayout()
        self.layout.padding_left = 20
        self.layout.padding_top = 20
        self.layout.padding_right = 20
        self.layout.padding_bottom = 20

        self.text_area = fsui.TextArea(self, read_only=True)
        self.text_area.set_min_width(600)
        self.text_area.set_min_height(300)
        self.layout.add(self.text_area)

        self.close_button = CloseButton.add_to_layout(
            self, self.layout, margin_top=20
        )
        if self.close_button:
            self.close_button.disable()

        self.line_count = 0
        self.task = ImportTask(path, import_type)
        self.task.start()

        self.set_size(self.layout.get_min_size())
        self.center_on_parent()
        fsui.call_later(TIMER_INTERVAL, self.on_timer)

    def on_timer(self):
        if self.task.done:
            LauncherSignal.broadcast("scan_done")
            if self.close_button:
                self.close_button.enable()
        else:
            fsui.call_later(TIMER_INTERVAL, self.on_timer)
        lines = self.task.get_new_log_lines(self.line_count)
        for line in lines:
            self.text_area.append_text(line + "\n")
        self.line_count += len(lines)



import fsui
from fsbc.application import app
from launcher.i18n import gettext
from launcher.launcher_settings import LauncherSettings
from launcher.option import Option
from launcher.settings.override_warning import OverrideWarning
from launcher.ui.HelpButton import HelpButton


class OptionUI(object):
    @classmethod
    def create_group(
        cls, parent, name, description=None, help_button=True, thin=False
    ):
        group = fsui.Group(parent)
        group.layout = fsui.HorizontalLayout()
        if thin:
            thin_layout = fsui.VerticalLayout()
            thin_layout.add(group.layout, fill=True)
        option = Option.get(name)
        if description == "":
            description = gettext(option["description"])
        if description:
            group.label = fsui.Label(group, description + ":")
            group.layout.add(group.label, margin_right=10)
            group.layout.add(OverrideWarning(group, name), margin_right=10)

        if thin:
            group.layout = fsui.HorizontalLayout()
            if description:
                thin_layout.add(group.layout, fill=True, margin_top=6)
            else:
                thin_layout.add(group.layout, fill=True, margin_top=0)

        choice_values = []

        if description:
            default_tmpl = "{0} (*)"
            # default_tmpl = "Default - {0}"
        else:
            default_tmpl = "{0} (*)"
            # default_tmpl = "Default - {0}"

        if option["type"].lower() == "boolean":
            if option["default"] == "1":
                default_desc = gettext(default_tmpl).format(gettext("On"))
            elif option["default"] == "0":
                default_desc = gettext(default_tmpl).format(gettext("Off"))
            else:
                default_desc = gettext("Default")
            choice_values.append(("", default_desc))
            choice_values.append(("1", gettext("On")))
            choice_values.append(("0", gettext("Off")))

        elif option["type"].lower() == "choice":
            for i, value in enumerate(option["values"]):
                if option["default"] == value[0]:
                    default_desc = gettext(default_tmpl).format(
                        gettext(value[1])
                    )
                    break
            else:
                default_desc = gettext("Default")
            choice_values.append(("", default_desc))
            for option in option["values"]:
                choice_values.append((option[0], gettext(option[1])))

        elif option["type"].lower() == "string":

            def on_changed():
                val = text_field.get_text()
                LauncherSettings.set(name, val.strip())

            text_field = fsui.TextField(group)
            # text_field.set_min_width(400)
            text_field.set_text(LauncherSettings.get(name))
            text_field.on_changed = on_changed
            group.layout.add(text_field, expand=True)

        elif (
            option["type"].lower() == "integer"
            and "min" in option
            and "max" in option
        ):
            current = LauncherSettings.get(name)

            if name == Option.LAUNCHER_FONT_SIZE:
                font = app.qapplication.font()
                Option.get(Option.LAUNCHER_FONT_SIZE)["default"] = (
                    font.pointSize()
                )

            current_int = int(option["default"])
            if current:
                try:
                    current_int = int(current)
                except ValueError:
                    pass
            current_int = max(option["min"], min(option["max"], current_int))
            check_box = fsui.CheckBox(group, gettext("Default"))
            spin_ctrl = fsui.SpinCtrl(
                group, option["min"], option["max"], current_int
            )
            if current == "":
                check_box.check()
                spin_ctrl.disable()

            def on_checkbox():
                if check_box.is_checked():
                    spin_ctrl.set_value(int(option["default"]))
                    spin_ctrl.disable()
                    LauncherSettings.set(name, "")
                else:
                    spin_ctrl.enable()

            check_box.on_changed = on_checkbox

            def on_spin():
                val = spin_ctrl.get_value()
                val = max(option["min"], min(option["max"], val))
                LauncherSettings.set(name, str(val))

            spin_ctrl.on_changed = on_spin
            group.layout.add_spacer(0, expand=True)
            group.layout.add(check_box)
            group.layout.add(spin_ctrl, margin_left=10)

        if choice_values:

            def on_changed():
                index = choice.get_index()
                LauncherSettings.set(name, choice_values[index][0])

            choice_labels = [x[1] for x in choice_values]
            choice = fsui.Choice(group, choice_labels)
            current = LauncherSettings.get(name)
            for i, value in enumerate(choice_values):
                if current == value[0]:
                    choice.set_index(i)
                    break
            choice.on_changed = on_changed
            if thin:
                group.layout.add(choice, expand=True)
            else:
                group.layout.add_spacer(0, expand=True)
                group.layout.add(choice)
            group.widget = choice

        if help_button:
            option_url = "https://fs-uae.net/docs/options/" + name.replace(
                "_", "-"
            )
            group.help_button = HelpButton(parent, option_url)
            group.layout.add(group.help_button, margin_left=10)

        if thin:
            group.layout = thin_layout
        return group



import signal
import sys

from fsui import Application
from fsui.qt import QFont
from fsui.qt import QFontDatabase
from workspace.ui import Font


class ApplicationRunner:
    def __init__(self, name):
        self._app = Application(name)

        stream = Font.stream("NotoSans-Regular.ttf")
        # noinspection PyArgumentList
        QFontDatabase.addApplicationFontFromData(stream.read())

        stream = Font.stream("Roboto-Regular.ttf")
        # noinspection PyArgumentList
        QFontDatabase.addApplicationFontFromData(stream.read())

        stream = Font.stream("RobotoMono-Regular.ttf")
        # noinspection PyArgumentList
        QFontDatabase.addApplicationFontFromData(stream.read())

        font = Font("Roboto", 14)
        font.qfont().setPointSizeF(10.5)
        font.qfont().setHintingPreference(QFont.PreferNoHinting)
        self._app.qapplication.setFont(font.qfont())

    def run(self, application):
        # FIXME: Find a nicer way to make PyQT cooperate with Ctrl+C?
        signal.signal(signal.SIGINT, signal.SIG_DFL)

        args = sys.argv[1:]
        application.init(args)
        # if args:
        #     print(args)
        #     application.arguments(args)

        self._app.run()



import fsui
from launcher.i18n import gettext


class PrevButton(fsui.Button):
    def __init__(self, parent):
        super().__init__(parent, gettext("Prev"))


class NextButton(fsui.Button):
    def __init__(self, parent):
        super().__init__(parent, gettext("Next"))


class CloseButton(fsui.Button):
    def __init__(self, parent):
        super().__init__(parent, gettext("Close"))

    def on_activate(self):
        print(repr(self.window))
        self.window.close()

    @classmethod
    def add_to_layout(cls, parent, layout, margin=0, margin_top=None):
        if parent.window.theme.has_close_buttons:
            button_layout = fsui.HorizontalLayout()
            assert isinstance(layout, fsui.Layout)
            layout.add(
                button_layout, fill=True, margin=margin, margin_top=margin_top
            )
            button_layout.add_spacer(expand=True)
            close_button = CloseButton(parent)
            button_layout.add(close_button, fill=True)
            return close_button
        else:
            return None



import fsui
from launcher.i18n import gettext
from launcher.settings.settings_page import SettingsPage


class NetplaySettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("netplay-settings", "pkg:workspace")
        gettext("Net Play Settings")
        title = gettext("Net Play")
        subtitle = ""
        self.add_header(icon, title, subtitle)

        self.add_option("irc_nick")
        self.add_option("netplay_tag")
        self.add_option("irc_server")



from fsgs.FSGameSystemContext import FSGameSystemContext
from fsgs.gamedb.constants import DATABASE_AMIGA


def game_database_dumper_main():
    fsgs = FSGameSystemContext()
    game_database = fsgs.game_database(DATABASE_AMIGA)
    # cursor = game_database.cursor()
    # cursor.execute("")
    for uuid in game_database.get_all_uuids():
        values = game_database.get_game_values_for_uuid(uuid, recursive=False)
        for key in sorted(values.keys()):
            print("{} = {}".format(key, values[key]))



from fsbc.desktop import open_url_in_browser
import workspace
import workspace.ui


class HelpButton(workspace.ui.ImageButton):
    def __init__(self, parent, url):
        super().__init__(
            parent,
            workspace.ui.Image(
                workspace.Stream(__name__, "data/16x16/information.png")
            ),
        )
        self.url = url
        self.set_min_width(40)

    def on_activate(self):
        open_url_in_browser(self.url)



import io
import os
from configparser import ConfigParser

from fsbc.resources import Resources
from fsbc.system import linux, macosx, windows
from fsbc.util import memoize
from fsgs.context import fsgs
from fsgs.FSGSDirectories import FSGSDirectories


class InputDeviceNotFoundException(Exception):
    pass


class MissingPlatformSupportException(Exception):
    pass


class InputDeviceInfo(object):
    id = ""
    generic_id = ""
    sdl_name = ""
    name = ""
    buttons = 0
    axes = 0
    hats = 0
    balls = 0
    index = 0

    def get_type(self):
        if self.sdl_name:
            return "TYPE_JOYSTICK"
        else:
            return "TYPE_KEYBOARD"


class InputDevice(object):
    MissingPlatformSupportException = MissingPlatformSupportException

    def __init__(
        self,
        platform,
        name,
        sclist,
        sdl_name="",
        sdl_joystick_id=-1,
        version=1,
        buttons=0,
        axes=0,
        hats=0,
        balls=0,
        multiple=True,
    ):
        """
        sclist -- system controller list
        """
        self.type = ""
        self.config = None
        self.config_inv = None
        self.id = name
        # self._name = name
        if "#" in self.id:
            self.name, dummy = self.id.rsplit("#", 1)
        else:
            self.name = self.id
        self.name = self.name.strip()
        self.decorate_name_with_number()

        # print("--------->", self.id)
        # print("------------>", self.name)
        # print("********** InputDevice Constructor", self.name, "ID is",
        #         self.id)

        self.platform = platform
        self.sdl_name = sdl_name
        if not self.sdl_name:
            self.sdl_name = self.id.rsplit("#", 1)[0].strip()
        self.sdl_joystick_id = sdl_joystick_id
        self.index = -1
        self.version = version
        self.buttons = buttons
        self.axes = axes
        self.hats = hats
        self.balls = balls
        for sc in sclist:
            # print("---", name, "---", sc.id, sc.index)
            if name == sc.id:
                self.index = sc.index
                break

        if version == 1:
            method_name = "configure_for_" + platform.lower().replace(
                " ", "_"
            ).replace("-", "_")
            if hasattr(self, method_name):
                getattr(self, method_name)(name, sclist=sclist)
            else:
                print(
                    "{0} does not support platform {1}".format(name, platform)
                )
                raise MissingPlatformSupportException(
                    "No input device support for platform"
                )
        elif version == 2:
            old_name = self.name
            self.configure_version_2(multiple=multiple)
            if self.name != old_name:
                # print("****** name is now", self.name)
                print(self.id, "=>", repr(self.name))

    def decorate_name_with_number(self):
        # print("decorate_name_with_number, ID is", self.id)
        if "#" in self.id:
            dummy, number_str = self.id.rsplit("#", 1)
        else:
            number_str = "1"
        if number_str != "1":
            self.name = "{0} #{1}".format(self.name, number_str)

    def is_joystick(self):
        return self.get_type() == "TYPE_JOYSTICK"

    def is_keyboard(self):
        return self.get_type() == "TYPE_KEYBOARD"

    def get_type(self):
        if self.type.startswith("joystick"):
            return "TYPE_JOYSTICK"
        elif self.type.startswith("keyboard"):
            return "TYPE_KEYBOARD"
        else:
            raise Exception("unknown input device type " + repr(self.type))

    @classmethod
    def get_builtin_config_for_device_guid(cls, guid):
        if windows:
            host_platform = "windows"
        elif macosx:
            host_platform = "macos"
        elif linux:
            host_platform = "linux"
        else:
            host_platform = "other"
        return Resources("fsgs").stream(
            "res/input/" + host_platform + "/" + guid + ".fs-uae-controller"
        )

    # @classmethod
    # def get_builtin_config_file_for_device_guid(cls, guid):
    #     try:
    #         path = Resources("fsgs").path(
    #             "res/input/" + guid + ".fs-uae-controller")
    #     except LookupError:
    #         return None
    #     return path

    @staticmethod
    @memoize
    def get_config_files():
        print("get_config_files")
        configs = {}
        input_stream = Resources("fsgs").stream("res/input/manifest.txt")
        print("opened input manifest")
        for line in input_stream.read().split(b"\n"):
            line = line.decode("UTF-8")
            path = line.strip()
            if not path:
                continue
            _, ext = os.path.splitext(path)
            if ext in [".ini", ".conf"]:
                parts = path.split("/")
                file_name = parts[-1]
                name, _ = os.path.splitext(file_name)
                if len(parts) > 1:
                    configs[parts[-2] + "_" + name] = "fsgs:res/input/" + path
                # print(" -", path)
                configs[name] = "fsgs:res/input/" + path

        # FIXME: fix dependency
        controllers_dir = fsgs.amiga.get_controllers_dir()
        print("read configs from controllers_dir at", controllers_dir)
        if os.path.exists(controllers_dir):
            for file_name in os.listdir(controllers_dir):
                if file_name.endswith(".conf"):
                    name, ext = os.path.splitext(file_name)
                    path = os.path.join(controllers_dir, file_name)
                    print(" -", path)
                    configs[name] = path
        keyboards_dir = os.path.join(
            FSGSDirectories.get_data_dir(), "Devs", "Keyboards"
        )
        print("read configs from controllers_dir at", keyboards_dir)
        if os.path.exists(keyboards_dir):
            for file_name in os.listdir(keyboards_dir):
                if file_name.endswith(".ini"):
                    name, ext = os.path.splitext(file_name)
                    path = os.path.join(keyboards_dir, file_name)
                    print(" -", path)
                    configs[name] = path

        # print("input config files:", configs)
        return configs

    def read_config(self, config_name, config, platform, multiple):
        print("read_config", config_name)
        configs = self.get_config_files()
        print(config_name in configs)
        try:
            path = configs[config_name]
        except KeyError:
            # traceback.print_stack()
            print(
                "no config file found for",
                repr(self.sdl_name),
                "=",
                config_name,
            )
            # raise InputDeviceNotFoundException(
            #         "no config found for " + repr(self.sdl_name))
            if platform:
                raise MissingPlatformSupportException(
                    "no config found for " + repr(self.sdl_name)
                )
            else:
                return
        cp = ConfigParser()
        print(path, os.path.exists(path))
        if path.startswith("fsgs:"):
            print("reading config from stream", path)
            input_stream = Resources("fsgs").stream(path.split(":", 1)[1])
            input_stream = io.TextIOWrapper(input_stream, "UTF-8")
            cp.read_file(input_stream)
        else:
            cp.read(path)
        if cp.has_option("device", "type"):
            self.type = cp.get("device", "type")
        if cp.has_option("device", "name"):
            # print("HAD NAME", self.name, self.sdl_name)
            self.name = cp.get("device", "name")
            self.decorate_name_with_number()
            # print("HAS NAME", self.name, self.sdl_name)
            # try:
            #     dummy, num = self.name.rsplit('#', 1)
            # except ValueError:
            #     self.name = name
            # else:
            #     self.name = "{0} #{1}".format(name, num)
        if cp.has_section(platform):
            section = platform
        elif cp.has_section("default"):
            # print("has default section")
            section = "default"
        else:
            if platform:
                raise MissingPlatformSupportException(
                    "no config found for platform " + repr(platform)
                )
            else:
                return
        # config = {}
        # if section == 'gamepad':
        #
        # for option in cp.options('gamepad'):
        #     value = cp.get('gamepad', option)
        #     print("gamepad", option, value)

        if cp.has_option(section, "include"):
            include_config = cp.get(section, "include")
            include_config = include_config.replace("/", "_")
            self.read_config(include_config, config, platform, multiple)

        # iconfig = {}
        # for key, value in config.items():
        #     iconfig[value] = key

        # for key in cp.options(section):
        #     value = cp.get(section, key)
        for key, value in cp.items(section):
            value = value.strip()
            if value.startswith("("):
                if not multiple:
                    continue
                assert value.endswith(")")
                value = value[1:-1]
            # print(key, "===>", value)
            # print("key, value is", key, value)
            # if value in iconfig:
            try:
                # config[key] = iconfig[value]
                # config[key] = config[value]
                # del config[value]
                # config[config[value]] = value

                # if key in iconfig:
                config[key] = config[value]
                # del iconfig[key]
                # iconfig[config[value]] = key
                # if not config[value] in iconfig:
                #     iconfig[config[value]] = key
                del config[value]
            except KeyError:
                config[key] = value
                # if not value in iconfig:
                #     iconfig[value] = key
                # config_order.append(key)

            # config[option] = cp.get(section, option)

        # if cp.has_section('gamepad'):
        #     for key, value in list(config.items()):
        #         if cp.has_option('gamepad', value):
        #             config[key] = cp.get('gamepad', value)

    def configure_version_2(self, multiple=True):
        self.config = self.configure(self.platform, multiple=multiple)

    def configure(self, platform, multiple=True):
        print("CONFIGURE", platform, self.name, self.sdl_name)
        # print("InputDevice.configure")

        name_lower = self.sdl_name.lower()
        name = ""
        for c in name_lower:
            if c in "abcdefghijklmnopqrstuvwxyz0123456789":
                name = name + c
            else:
                if not name.endswith("_"):
                    name += "_"
        name = name.strip("_")
        if windows:
            host_platform = "windows"
        elif macosx:
            host_platform = "macos"
        elif linux:
            host_platform = "linux"
        else:
            host_platform = "other"
        config_name = "{0}_{1}_{2}_{3}_{4}_{5}".format(
            name, self.buttons, self.axes, self.hats, self.balls, host_platform
        )

        config = {}
        try:
            self.read_config(self.name, config, platform, multiple)
        except Exception:
            pass
        else:
            return config
        try:
            self.read_config(self.sdl_name, config, platform, multiple)
        except Exception:
            pass
        else:
            return config
        try:
            self.read_config(config_name, config, platform, multiple)
        except Exception:
            pass
        else:
            return config

        config_name = ""
        for c in self.sdl_name.lower():
            if c in "abcdefghijklmnopqrstuvwxyz0123456789":
                config_name += c
            elif len(config_name) == 0 or config_name[-1] != "_":
                config_name += "_"
        if config_name.endswith("_usb"):
            config_name = config_name[:-4]
        while config_name.endswith("_"):
            config_name = config_name[:-1]
        # print("config_name =", config_name, "sdl_name", repr(self.sdl_name))
        self.read_config(config_name, config, platform, multiple)
        return config

    @staticmethod
    def supported_keyboard_input_devices():
        return "", 0

    def get_name(self):
        return self.name

    def get_sdl_joystick_id(self):
        return self.sdl_joystick_id

    def get_config(self):
        if self.config is None:
            raise Exception("InputDevice is not configured")
        return self.config

    def get_config_inverted(self):
        if self.config_inv is not None:
            return self.config_inv
        if self.config is None:
            raise Exception("InputDevice is not configured")
        self.config_inv = {}
        for k, v in self.config.items():
            self.config_inv[v] = k
        return self.config_inv



import sys

from fsbc.settings import Settings
from fsgs.context import default_context


def app_main():
    args = []
    for arg in sys.argv[1:]:
        if not arg.startswith("--"):
            args.append(arg)
    if len(args) == 0:
        print("Usage: fsgc [options] <game_uuid | variant_uuid>")
        return

    game_or_variant_uuid = args[-1]
    print(game_or_variant_uuid)

    fsgc = default_context()
    fsgc.load_game_by_uuid(game_or_variant_uuid)
    fsgc.config.add_from_argv()

    print("settings:fullscreen", Settings.instance()["fullscreen"])
    print("config:fullscreen", fsgc.config.get("fullscreen"))

    # sys.exit(1)
    # fsgc.run_game()

    from fsgs.platform import Platform

    platform = Platform.create(fsgc.game.platform.id)
    driver = platform.driver(fsgc)

    from fsgs.input.enumeratehelper import EnumerateHelper

    device_helper = EnumerateHelper()
    device_helper.default_port_selection(driver.ports, driver.options)

    print("")
    for port in driver.ports:
        print("Port", port.index)
        print(" ", port.type, [x["type"] for x in port.types])
        print(" ", port.device)
    print("")

    # sys.exit(1)

    driver.prepare()
    process = driver.run()
    process.wait()
    driver.finish()



import os

from fsgs.util.gamenameutil import GameNameUtil


class DiskUtil(object):
    @classmethod
    def find_disks(cls, path, base_names=None, script=None):
        disks = [path]
        print("")
        print("FIRST DISK", path)
        print("")
        # disks = []
        use_name = script.rename_file(os.path.basename(path))
        number = GameNameUtil.find_number(use_name)
        first_without_number = GameNameUtil.strip_number(use_name)
        first_without_flags = GameNameUtil.strip_flags(first_without_number)
        if number is None:
            # only this disk
            return disks
        base_name = GameNameUtil.find_base_name(path)
        dir = os.path.dirname(path)
        candidates = {}
        items = []
        if base_names:
            items = base_names[base_name]
        else:
            all_items = os.listdir(dir)
            for item in all_items:
                b = GameNameUtil.find_base_name(item)
                if base_name == b:
                    items.append(item)
        # print("items is", items)
        for item in items:
            use_name = script.rename_file(item)
            if GameNameUtil.is_bad_dump(use_name):
                continue
            n = GameNameUtil.find_number(use_name)
            if n == 0:
                n = 99
            if n:
                candidates.setdefault(n, []).append(
                    (use_name, os.path.join(dir, item))
                )
        # print(candidates)
        for n, items in candidates.items():
            print("")
            print("FIND FLOPPIES FOR n", n, "=", items)
            print("")
            if n == 1:
                # already added first floppy
                continue
            print(" - floppy number", n)
            matches = []
            for use_name, p in items:
                without_number = GameNameUtil.strip_number(use_name)
                # print(without_number, "vs", first_without_number)
                if without_number == first_without_number:
                    print("perfect match")
                    # disks.append(p)
                    matches.append((0, p))
                    continue
                    # break
                without_flags = GameNameUtil.strip_flags(without_number)
                if without_flags == first_without_flags:
                    print("ok match")
                    if without_flags == without_number:
                        # there were no flags on this floppy
                        matches.append((1, p))
                    else:
                        # slightly worse score since there were flags
                        # not matching floppy 1
                        matches.append((2, p))
                        # disks.append(p)
                        # break
                        # matches.append((2, p))
            if len(matches) == 0:
                # raise Exception("no candidates for floppy " + num)
                # print("WARNING: choosing partial matching floppy", p,
                #         "for floppy number", n)
                # disks.append(p)
                print("Did not find good match for floppy", n)
                print("candidates:")
                for item in items:
                    print("  ", item)
                raise Exception(
                    "Did not find good match for floppy {0}".format(n)
                )
            matches.sort()
            print(matches)
            score, p = matches[0]
            # if score == 2:
            disks.append(p)
            # TOSEC (x of y) disk number labelling format
            if " of {0})".format(n) in path:
                # found the correct number of disks
                break
        return disks



import sys

from .device import Device
from .enumeratehelper import EnumerateHelper


class DeviceManager(object):
    __instance = None

    @classmethod
    def instance(cls):
        if cls.__instance is None:
            helper = EnumerateHelper()
            cls.__instance = cls(helper)
        return cls.__instance

    def __init__(self, helper):
        self.helper = helper
        self.devices = []
        if "--add-dummy-joystick" in sys.argv:
            self.add_joystick_device(
                {
                    "axes": 2,
                    "hats": 0,
                    "balls": 0,
                    "buttons": 1,
                    "name": "Dummy Joystick",
                    "id": "Dummy Joystick",
                }
            )

    def get_devices(self):
        print("DeviceManager.get_devices")
        if self.helper is not None:
            self.helper.init()
            return self.helper.devices
        print("DeviceManager.get_devices ->", self.devices)
        return self.devices

    def add_device_from_event(self, event):
        if event["type"] == "joy-device-added":
            return self.add_joystick_device(event)
        elif event["type"] == "mouse-device-added":
            return self.add_mouse_device(event)
        elif event["type"] == "keyboard-device-added":
            return self.add_keyboard_device(event)
        else:
            assert False

    def add_joystick_device(self, event):
        device = Device()
        device.axes = event["axes"]
        device.balls = event["balls"]
        device.hats = event["hats"]
        device.buttons = event["buttons"]
        device.name = event["name"]
        device.id = event["id"]
        device.type = Device.TYPE_JOYSTICK
        self.devices.append(device)
        print("[INPUT] Joystick device added:", device.name)
        return device

    def add_keyboard_device(self, event):
        device = Device()
        device.name = event["name"]
        device.id = event["id"]
        device.type = Device.TYPE_KEYBOARD
        self.devices.append(device)
        print("[INPUT] Keyboard device added:", device.name)
        return device

    def add_mouse_device(self, event):
        device = Device()
        device.name = event["name"]
        device.id = event["id"]
        device.type = Device.TYPE_MOUSE
        self.devices.append(device)
        print("[INPUT] Mouse device added:", device.name)
        return device

    def remove_all_devices(self):
        for device in self.devices:
            print("[INPUT] Removing device", device.name)
        self.devices.clear()



from __future__ import absolute_import
from __future__ import print_function

import time
import struct

from .Block import Block
from ..ProtectFlags import ProtectFlags
from ..TimeStamp import TimeStamp

class DirCacheRecord:
  def __init__(self, entry=0, size=0, protect=0, mod_ts=None, sub_type=0, name='', comment=''):
    self.entry = entry
    self.size = size
    self.protect = protect
    self.mod_ts = mod_ts
    self.sub_type = sub_type
    self.name = name
    if comment == None:
      self.comment = ''
    else:
      self.comment = comment
    self.offset = None
  
  def get_size(self):
    total_len = 25 + len(self.name) + len(self.comment)
    align_len = (total_len + 1) & ~1
    return align_len
    
  def get(self, data, off):
    self.offset = off
    # header
    d = struct.unpack_from(">IIIHHHHH",data,offset=off)
    self.entry = d[0]
    self.size = d[1]
    self.protect = d[2]
    self.mod_ts = TimeStamp(d[5],d[6],d[7])
    self.type = ord(data[off + 22])
    # name
    name_len = ord(data[off + 23])
    name_off = off + 24
    self.name = data[name_off : name_off + name_len]
    # comment
    comment_len = ord(data[off + name_len + 24])
    comment_off = off + 25 + name_len
    self.comment = data[comment_off : comment_off + comment_len]
    return off + self.get_size()
    
  def put(self, data, off):
    self.offset = off
    # header
    ts = self.mod_ts
    struct.pack_into(">IIIHHHHH",data,off,self.entry,self.size,self.protect,0,0,ts.days,ts.mins,ts.ticks)
    # name
    name_len = len(self.name)
    data[off + 23] = chr(name_len)
    name_off = off + 24
    data[name_off : name_off + name_len] = self.name
    # comment
    comment_len = len(self.comment)
    data[off + 24 + name_len] = chr(comment_len)
    comment_off = off + 25 + name_len
    data[comment_off : comment_off + comment_len] = self.comment
    return off + self.get_size()
    
  def dump(self):
    print("DirCacheRecord(%s)(size=%d)" % (self.offset, self.get_size()))
    print("\tentry:      %s" % self.entry)
    print("\tsize:       %s" % self.size)
    pf = ProtectFlags(self.protect)
    print("\tprotect:    0x%x 0b%s %s" % (self.protect, pf.bin_str(), pf))
    print("\tmod_ts:     %s" % self.mod_ts)
    print("\tsub_type:   0x%x" % self.sub_type)
    print("\tname:       %s" % self.name)
    print("\tcomment:    %s" % self.comment)

class DirCacheBlock(Block):
  def __init__(self, blkdev, blk_num):
    Block.__init__(self, blkdev, blk_num, is_type=Block.T_DIR_CACHE)
  
  def set(self, data):
    self._set_data(data)
    self._read()
  
  def read(self):
    self._read_data()
    self._read()
  
  def _read(self):
    Block.read(self)
    if not self.valid:
      return False
    
    # fields
    self.own_key = self._get_long(1)
    self.parent = self._get_long(2)
    self.num_records = self._get_long(3)
    self.next_cache = self._get_long(4)  
    self.records = []

    # get records
    off = 24
    for i in xrange(self.num_records):
      r = DirCacheRecord()
      off = r.get(self.data, off)
      if off == -1:
        return False
      self.records.append(r)
    
    self.valid = True
    return True
    
  def get_total_record_size(self):
    size = 0
    for r in self.records:
      size += r.get_size()
    return size
    
  def get_free_record_size(self):
    return self.blkdev.block_bytes - 24 - self.get_total_record_size()

  def create(self, parent, records=None, next_cache=0):
    Block.create(self)
    self.own_key = self.blk_num
    self.parent = parent
    self.next_cache = next_cache
    if records == None:
      self.num_records = 0
      self.records = []
    else:
      self.num_records = len(records)
      self.records = records
    self.valid = True
    return True
  
  def add_record(self, record):
    self.records.append(record)
    self.num_records = len(self.records)
  
  def get_record_by_name(self, name):
    for r in self.records:
      if r.name == name:
        return r
    return None
  
  def remove_record(self, record):
    self.records.remove(record)
    self.num_records = len(self.records)
  
  def has_record(self, record):
    return record in self.records
  
  def is_empty(self):
    return self.num_records == 0
  
  def write(self):
    Block._create_data(self)
    self._put_long(1, self.own_key)
    self._put_long(2, self.parent)
    self._put_long(3, self.num_records)
    self._put_long(4, self.next_cache)
    
    # put records
    off = 24
    for r in self.records:
      off = r.put(self.data, off)
    
    Block.write(self)
  
  def dump(self):
    Block.dump(self,"DirCache")
    print(" own_key:    %d" % (self.own_key))
    print(" parent:     %d" % (self.parent))
    print(" num_records:%d" % (self.num_records))
    print(" next_cache: %d" % (self.next_cache))
    for r in self.records:
      r.dump()



from .irc_broadcaster import IRCBroadcaster
from .irc_color import IRCColor


class Channel:
    def __init__(self, irc, name):
        self.name = name
        self.lines = []
        self.colors = []
        self.nicks = set()
        self.ops = set()
        self.half_ops = set()
        self.voices = set()
        self.irc = irc

    def get_text(self):
        return "\n".join(self.lines) + "\n"

    def is_op(self, nick=None):
        if nick is None:
            nick = self.irc.my_nick
        return nick in self.ops

    def message(self, message, color=None):
        if color is None:
            # default color does not work properly on Mac
            color = IRCColor.MESSAGE
        # remove special character (action)
        message = message.replace("\u0001", "")
        # remove special character (bold font)
        message = message.replace("\u0002", "")
        line = message + "\n"
        self.lines.append(line)
        self.colors.append(color)
        IRCBroadcaster.broadcast(
            "message", {"channel": self.name, "message": line, "color": color}
        )

    def info(self, message):
        return self.message(message, color=IRCColor.INFO)

    def warning(self, message):
        return self.message(message, color=IRCColor.WARNING)

    def get_nick_list(self):
        nicks = []
        for nick in self.nicks:
            if nick in self.ops:
                status = "@"
            elif nick in self.half_ops:
                status = "%"
            elif nick in self.voices:
                status = "+"
            else:
                status = ""
            nicks.append(status + nick)
        return sorted(nicks)

    def action(self, message):
        message = "\u0001ACTION {0}\u0001".format(message)
        self.privmsg(message)

    def privmsg(self, message, notice=False, echo=True):
        if echo and not message.startswith("__"):
            if notice:
                text = "-{0}- {1}".format(self.irc.my_nick, message)
            elif message.startswith("[*] "):
                text = "* {0} {1}".format(self.irc.my_nick, message[4:])
            elif message.startswith("\u0001ACTION"):
                text = "* {0} {1}".format(self.irc.my_nick, message[8:-1])
            else:
                text = "<{0}> {1}".format(self.irc.my_nick, message)
            self.message(text, IRCColor.MY_MESSAGE)
        if notice:
            self.irc.notice(self.name, message)
        else:
            self.irc.privmsg(self.name, message)

    def notice(self, message):
        return self.privmsg(message, notice=True)

    def add_nick(self, nick):
        self.nicks.add(nick)
        IRCBroadcaster.broadcast("join", {"channel": self.name, "nick": nick})

    def remove_nick(self, nick):
        IRCBroadcaster.broadcast("part", {"channel": self.name, "nick": nick})
        if nick not in self.nicks:
            # print("Channel.parted - warning, nick not in list", nick)
            pass
        else:
            self.nicks.remove(nick)
            IRCBroadcaster.broadcast("nick_list", {"channel": self.name})
        if nick in self.ops:
            self.ops.remove(nick)
        if nick in self.half_ops:
            self.half_ops.remove(nick)
        if nick in self.voices:
            self.voices.remove(nick)

    def handle_leave_channel(self):
        self.nicks.clear()
        self.ops.clear()
        self.voices.clear()
        IRCBroadcaster.broadcast("parted", {"channel": self.name})

    def on_privmsg(self, nick, message, notice=False):
        if message.startswith("__"):
            pass
        else:
            color = IRCColor.MESSAGE
            if notice:
                text = "-{0}- {1}".format(nick, message)
                color = IRCColor.NOTICE
            elif message.startswith("[*] "):
                text = "* {0} {1}".format(nick, message[4:])
            elif message.startswith("\u0001ACTION"):
                text = "* {0} {1}".format(nick, message[8:-1])
            else:
                text = "<{0}> {1}".format(nick, message)
            self.message(text, color)

    def on_notice(self, nick, message):
        return self.on_privmsg(nick, message, notice=True)

    # noinspection SpellCheckingInspection
    def on_currenttopic(self, topic):
        self.message("{0}".format(topic), IRCColor.TOPIC)

    def on_join(self, nick):
        if self.irc.me(nick):
            self.nicks.clear()
            self.ops.clear()
            self.voices.clear()
            self.message("* you joined {0} ".format(self.name), IRCColor.JOIN)
            self.irc.set_active_channel_name(self.name)
            IRCBroadcaster.broadcast("joined", {"channel": self.name})
        else:
            self.message(
                "* {0} joined ({1}) ".format(nick, self.name), IRCColor.JOIN
            )
            if nick in self.nicks:
                print("Channel.joined - warning, nick already in list", nick)
            else:
                # self.nicks.add(nick)
                self.add_nick(nick)
        IRCBroadcaster.broadcast("nick_list", {"channel": self.name})

    # noinspection SpellCheckingInspection
    def on_kick(self, kicker, kickee, reason):
        self.message(
            "* {0} kicked {1} ({2})".format(kicker, kickee, reason),
            IRCColor.KICK,
        )
        if self.irc.me(kickee):
            self.handle_leave_channel()

    def on_topic(self, who, topic):
        self.message(
            "* {0} changed topic to:\n{1} ".format(who, topic), IRCColor.TOPIC
        )

    def on_mode(self, who, args):
        if args[0] == "+o":
            for nick in args[1:]:
                self.ops.add(nick)
        elif args[0] == "-o":
            for nick in args[1:]:
                if nick in self.ops:
                    self.ops.remove(nick)
        elif args[0] == "+h":
            for nick in args[1:]:
                self.half_ops.add(nick)
        elif args[0] == "-h":
            for nick in args[1:]:
                if nick in self.half_ops:
                    self.half_ops.remove(nick)
        elif args[0] == "+v":
            for nick in args[1:]:
                self.voices.add(nick)
        elif args[0] == "-v":
            for nick in args[1:]:
                if nick in self.voices:
                    self.voices.remove(nick)
        if args[0].startswith("+"):
            color = IRCColor.POS_MODE
        else:
            color = IRCColor.NEG_MODE
        self.message("* {0} sets mode {1}".format(who, " ".join(args)), color)
        IRCBroadcaster.broadcast("nick_list", {"channel": self.name})

    # noinspection SpellCheckingInspection
    def on_namreply(self, nicks):
        for nick in nicks:
            if nick.startswith("@"):
                # self.nicks.add(nick[1:])
                self.ops.add(nick[1:])
                self.add_nick(nick[1:])
            elif nick.startswith("+"):
                # self.nicks.add(nick[1:])
                self.voices.add(nick[1:])
                self.add_nick(nick[1:])
            else:
                # self.nicks.add(nick)
                self.add_nick(nick)
        IRCBroadcaster.broadcast("nick_list", {"channel": self.name})

    def on_part(self, nick):
        if self.irc.me(nick):
            self.message("* you left " + self.name, IRCColor.PART)
            self.handle_leave_channel()
        else:
            self.message(
                "* {0} left {1}".format(nick, self.name), IRCColor.PART
            )
            self.remove_nick(nick)
        IRCBroadcaster.broadcast("nick_list", {"channel": self.name})

    def on_quit(self, nick, reason):
        if nick in self.nicks or self.name == nick or not self.name:
            self.message(
                "* {0} quit ({1}) ".format(nick, reason), IRCColor.PART
            )
        self.remove_nick(nick)



import launcher.version
from arcade.Application import Application
from arcade.glui.imageloader import ImageLoader
from arcade.ui.arcade_window import (
    ArcadeWindow,
    check_argument,
    fullscreen,
    maximized,
)
from fsbc.init import initialize_application
from fsbc.settings import Settings
from fsbc.system import macosx

from .gnome3 import handle_gnome_extensions, running_in_gnome_3

K_UI_MODE_ALL_HIDDEN = 3
K_UI_OPTION_AUTO_SHOW_MENU_BAR = 1 << 0


def os_x_set_system_ui_mode(mode, option):
    # noinspection PyUnresolvedReferences
    import objc

    # noinspection PyUnresolvedReferences
    from Foundation import NSBundle

    bundle = NSBundle.bundleWithPath_(
        "/System/Library/Frameworks/Carbon.framework"
    )
    objc.loadBundleFunctions(
        bundle, globals(), (("SetSystemUIMode", b"III", ""),)
    )
    # noinspection PyUnresolvedReferences
    SetSystemUIMode(mode, option)


def main():
    application = Application()
    initialize_application("fs-uae-arcade", version=launcher.version.VERSION)

    # fs_width, fs_height = fsui.get_screen_size()
    # cursor_position = None

    # use_window = False
    # use_window_decorations = True
    # use_fullscreen = True
    # use_fullscreen_window = False
    # use_top_clock = check_argument("top_clock") != "0"
    # use_top_logo = check_argument("top_logo") != "0"

    if macosx:
        if fullscreen() or maximized():
            if check_argument("system_autohide") == "1":
                os_x_set_system_ui_mode(
                    K_UI_MODE_ALL_HIDDEN, K_UI_OPTION_AUTO_SHOW_MENU_BAR
                )
    elif running_in_gnome_3():
        if fullscreen() or maximized():
            # use_fullscreen = False
            # use_window_decorations = False
            # use_window = "maximized"
            if check_argument("system_autohide") == "1":
                handle_gnome_extensions()
                # cursor_position = fs_width - 1, fs_height - 1
                # use_top_clock = False
                # use_top_logo = False
                # app.settings["fs-uae:fullscreen-mode::default"] = "window"
        else:
            # We want a normal window.
            pass

    Settings.instance().set("__arcade", "1")

    # if windows:
    #     pass
    # elif macosx:
    #     # use_fullscreen_window = True
    #     # Settings.instance().set("__fullscreen_mode", "window")
    #     pass
    # else:
    #     # app.settings["fs-uae:fullscreen-mode::default"] = "window"
    #     pass

    # if check_argument("fullscreen"):
    #     use_fullscreen = check_argument("fullscreen") == "1"
    #
    # if "--fullscreen-mode=fullscreen" in sys.argv:
    #     use_fullscreen_window = False
    # elif "--fullscreen-mode=window" in sys.argv:
    #     use_fullscreen_window = True
    #
    # if "--maximize" in sys.argv:
    #     use_window = "maximized"
    #     use_fullscreen = False
    #
    # if "--no-window-decorations" in sys.argv:
    #     use_window_decorations = False

    # app.settings["game-center:fullscreen"] = \
    #     "1" if use_fullscreen else "0"
    # if use_fullscreen_window:
    #     app.settings["game-center:fullscreen-mode"] = "window"
    # else:
    #     app.settings["game-center:fullscreen-mode"] = ""
    # app.settings["game-center:window-decorations"] = \
    #     "1" if use_window_decorations else "0"
    # app.settings["game-center:maximize"] = \
    #     "1" if use_window == "maximized" else "0"
    # app.settings["game-center:top-clock"] = "1" if use_top_clock else "0"
    # app.settings["game-center:top-logo"] = "1" if use_top_logo else "0"

    ArcadeWindow().show_auto()

    # if cursor_position is not None:
    #     os.environ["FSGS_RETURN_CURSOR_TO"] = "{0},{1}".format(
    #         cursor_position[0], cursor_position[1])

    application.run()
    print("application.run returned")

    application.stop()
    ImageLoader.get().stop()

    application.wait()

    print(" --- arcade.arcade_main.main is done ---")
    return



from __future__ import absolute_import
from __future__ import print_function

from amitools.fs.block.rdb.RDBlock import *
from amitools.fs.block.rdb.PartitionBlock import *
import amitools.util.ByteSize as ByteSize
import amitools.fs.DosType as DosType
from .FileSystem import FileSystem
from .Partition import Partition

class RDisk:
  def __init__(self, rawblk):
    self.rawblk = rawblk
    self.valid = False
    self.rdb = None
    self.parts = []
    self.fs = []
    self.used_blks = []
    self.max_blks = 0

  def open(self):
    # read RDB
    self.rdb = RDBlock(self.rawblk)
    if not self.rdb.read():
      self.valid = False
      return False
    # create used block list
    self.used_blks = [self.rdb.blk_num]

    # read partitions
    part_blk = self.rdb.part_list
    self.parts = []
    num = 0
    while part_blk != Block.no_blk:
      p = Partition(self.rawblk, part_blk, num, self.rdb.log_drv.cyl_blks, self)
      num += 1
      if not p.read():
        self.valid = False
        return False
      self.parts.append(p)
      # store used block
      self.used_blks.append(p.get_blk_num())
      # next partition
      part_blk = p.get_next_partition_blk()

    # read filesystems
    fs_blk = self.rdb.fs_list
    self.fs = []
    num = 0
    while fs_blk != PartitionBlock.no_blk:
      fs = FileSystem(self.rawblk, fs_blk, num)
      num += 1
      if not fs.read():
        self.valid = False
        return False
      self.fs.append(fs)
      # store used blocks
      self.used_blks += fs.get_blk_nums()
      # next partition
      fs_blk = fs.get_next_fs_blk()

    # TODO: add bad block blocks

    self.valid = True
    self.max_blks = self.rdb.log_drv.rdb_blk_hi + 1
    return True

  def close(self):
    pass

  # ----- query -----

  def dump(self, hex_dump=False):
    # rdb
    if self.rdb != None:
      self.rdb.dump()
    # partitions
    for p in self.parts:
      p.dump()
    # fs
    for fs in self.fs:
      fs.dump(hex_dump)

  def get_info(self):
    res = []
    # physical disk info
    pd = self.rdb.phy_drv
    total_blks = self.get_total_blocks()
    total_bytes = self.get_total_bytes()
    extra="heads=%d sectors=%d" % (pd.heads, pd.secs)
    res.append("PhysicalDisk:        %8d %8d  %10d  %s  %s" \
      % (0, pd.cyls-1, total_blks, ByteSize.to_byte_size_str(total_bytes), extra))
    # logical disk info
    ld = self.rdb.log_drv
    extra="rdb_blks=[%d:%d,#%d] used=[hi=%d,#%d] cyl_blks=%d" % (ld.rdb_blk_lo, ld.rdb_blk_hi, self.max_blks, ld.high_rdsk_blk, len(self.used_blks), ld.cyl_blks)
    logic_blks = self.get_logical_blocks()
    logic_bytes = self.get_logical_bytes()
    res.append("LogicalDisk:         %8d %8d  %10d  %s  %s" \
      % (ld.lo_cyl, ld.hi_cyl, logic_blks, ByteSize.to_byte_size_str(logic_bytes), extra))
    # add partitions
    for p in self.parts:
      res.append(p.get_info(logic_blks))
    # add fileystems
    for f in self.fs:
      res.append(f.get_info())
    return res

  def get_block_map(self):
    res = []
    for i in xrange(self.max_blks):
      blk = None
      # check partitions
      if i == 0:
        blk = "RD"
      else:
        for p in self.parts:
          if i == p.get_blk_num():
            blk = 'P%d' % p.num
            break
        if blk == None:
          # check file systems
          for f in self.fs:
            if i in f.get_blk_nums():
              blk = 'F%d' % f.num
              break
        if blk == None:
          blk = '--'
      res.append(blk)
    return res

  def get_logical_cylinders(self):
    ld = self.rdb.log_drv
    return ld.hi_cyl - ld.lo_cyl + 1

  def get_logical_blocks(self):
    ld = self.rdb.log_drv
    cyls = ld.hi_cyl - ld.lo_cyl + 1
    return cyls * ld.cyl_blks

  def get_logical_bytes(self, block_bytes=512):
    return self.get_logical_blocks() * block_bytes

  def get_total_blocks(self):
    pd = self.rdb.phy_drv
    return pd.cyls * pd.heads * pd.secs

  def get_total_bytes(self, block_bytes=512):
    return self.get_total_blocks() * block_bytes

  def get_cylinder_blocks(self):
    ld = self.rdb.log_drv
    return ld.cyl_blks

  def get_cylinder_bytes(self, block_bytes=512):
    return self.get_cylinder_blocks() * block_bytes

  def get_num_partitions(self):
    return len(self.parts)

  def get_partition(self, num):
    if num < len(self.parts):
      return self.parts[num]
    else:
      return None

  def find_partition_by_drive_name(self, name):
    lo_name = name.lower()
    num = 0
    for p in self.parts:
      drv_name = p.get_drive_name().lower()
      if drv_name == lo_name:
        return p
    return None

  def find_partition_by_string(self, s):
    p = self.find_partition_by_drive_name(s)
    if p != None:
      return p
    # try partition number
    try:
      num = int(s)
      return self.get_partition(num)
    except ValueError:
      return None

  def get_filesystem(self, num):
    if num < len(self.fs):
      return self.fs[num]
    else:
      return None

  def get_used_blocks(self):
    return self.used_blks

  def find_filesystem_by_string(self, s):
    # try filesystem number
    try:
      num = int(s)
      return self.get_filesystem(num)
    except ValueError:
      return None

  # ----- edit -----

  def create(self, disk_geo, rdb_cyls=1, hi_rdb_blk=0, disk_names=None, ctrl_names=None):
    cyls = disk_geo.cyls
    heads = disk_geo.heads
    secs = disk_geo.secs
    cyl_blks = heads * secs
    rdb_blk_hi = cyl_blks * rdb_cyls - 1

    if disk_names != None:
      disk_vendor = disk_names[0]
      disk_product = disk_names[1]
      disk_revision = disk_names[2]
    else:
      disk_vendor = 'RDBTOOL'
      disk_product = 'IMAGE'
      disk_revision = '2012'

    if ctrl_names != None:
      ctrl_vendor = ctrl_names[0]
      ctrl_product = ctrl_names[1]
      ctrl_revision = ctrl_names[2]
    else:
      ctrl_vendor = ''
      ctrl_product = ''
      ctrl_revision = ''

    flags = 0x7
    if disk_names != None:
      flags |= 0x10
    if ctrl_names != None:
      flags |= 0x20

    # create RDB
    phy_drv = RDBPhysicalDrive(cyls, heads, secs)
    log_drv = RDBLogicalDrive(rdb_blk_hi=rdb_blk_hi, lo_cyl=rdb_cyls, hi_cyl=cyls-1, cyl_blks=cyl_blks, high_rdsk_blk=hi_rdb_blk)
    drv_id = RDBDriveID(disk_vendor, disk_product, disk_revision, ctrl_vendor, ctrl_product, ctrl_revision)
    self.rdb = RDBlock(self.rawblk)
    self.rdb.create(phy_drv, log_drv, drv_id, flags=flags)
    self.rdb.write()

    self.used_blks = [self.rdb.blk_num]
    self.max_blks = self.rdb.log_drv.rdb_blk_hi + 1
    self.valid = True

  def get_cyl_range(self):
    log_drv = self.rdb.log_drv
    return (log_drv.lo_cyl, log_drv.hi_cyl)

  def check_cyl_range(self, lo_cyl, hi_cyl):
    if lo_cyl > hi_cyl:
      return False
    (lo,hi) = self.get_cyl_range()
    if not (lo_cyl >= lo and hi_cyl <= hi):
      return False
    # check partitions
    for p in self.parts:
      (lo,hi) = p.get_cyl_range()
      if not ((hi_cyl < lo) or (lo_cyl > hi)):
        return False
    return True

  def get_free_cyl_ranges(self):
    lohi = self.get_cyl_range()
    free = [lohi]
    for p in self.parts:
      pr = p.get_cyl_range()
      new_free = []
      for r in free:
        # partition completely fills range
        if pr[0] == r[0] and pr[1] == r[1]:
          pass
        # partition starts at range
        elif pr[0] == r[0]:
          n = (pr[1]+1, r[1])
          new_free.append(n)
        # partition ends at range
        elif pr[1] == r[1]:
          n = (r[0], pr[0]-1)
          new_free.append(n)
        # partition inside range
        elif pr[0] > r[0] and pr[1] < r[1]:
          new_free.append((r[0], pr[0]-1))
          new_free.append((pr[1]+1,r[1]))
        else:
          new_free.append(r)
      free = new_free
    if len(free) == 0:
      return None
    return free

  def find_free_cyl_range_start(self, num_cyls):
    ranges = self.get_free_cyl_ranges()
    if ranges == None:
      return None
    for r in ranges:
      size = r[1] - r[0] + 1
      if num_cyls <= size:
        return r[0]
    return None

  # ----- manage rdb blocks -----

  def _has_free_rdb_blocks(self, num):
    return (len(self.used_blks) + num) <= self.max_blks

  def get_free_blocks(self):
    res = []
    for i in xrange(self.max_blks):
      if i not in self.used_blks:
        res.append(i)
    return res

  def _alloc_rdb_blocks(self, num):
    free = self.get_free_blocks()
    n = len(free)
    if n == num:
      return free
    elif n > num:
      return free[:num]
    else:
      return None

  def _next_rdb_block(self):
    free = self.get_free_blocks()
    if len(free) > 0:
      return free[0]
    else:
      return None

  def _update_hi_blk(self):
    hi = 0
    for blk_num in self.used_blks:
      if blk_num > hi:
        hi = blk_num
    ld = self.rdb.log_drv
    ld.high_rdsk_blk = hi

  # ----- partition handling -----

  def _adjust_dos_env(self, dos_env, more_dos_env):
    if more_dos_env is None:
      return
    for p in more_dos_env:
      key = p[0]
      if hasattr(dos_env, key):
        setattr(dos_env, key, int(p[1]))

  def add_partition(self, drv_name, cyl_range, dev_flags=0, flags=0, dos_type=DosType.DOS0, boot_pri=0, more_dos_env=None):
    # cyl range is not free anymore or invalid
    if not self.check_cyl_range(*cyl_range):
      return False
    # no space left for partition block
    if not self._has_free_rdb_blocks(1):
      return False
    # allocate block for partition
    blk_num = self._alloc_rdb_blocks(1)[0]
    self.used_blks.append(blk_num)
    self._update_hi_blk()
    # crete a new parttion block
    pb = PartitionBlock(self.rawblk, blk_num)
    # setup dos env
    heads = self.rdb.phy_drv.heads
    blk_per_trk = self.rdb.phy_drv.secs
    dos_env = PartitionDosEnv(low_cyl=cyl_range[0], high_cyl=cyl_range[1], surfaces=heads, \
                              blk_per_trk=blk_per_trk, dos_type=dos_type, boot_pri=boot_pri)
    self._adjust_dos_env(dos_env, more_dos_env)
    pb.create(drv_name, dos_env, flags=flags)
    pb.write()
    # link block
    if len(self.parts) == 0:
      # write into RDB
      self.rdb.part_list = blk_num
    else:
      # write into last partition block
      last_pb = self.parts[-1]
      last_pb.part_blk.next = blk_num
      last_pb.write()
    # always write RDB as allocated block is stored there, too
    self.rdb.write()
    # flush out all changes before we read again
    self.rawblk.flush()
    # create partition object and add to partition list
    blk_per_cyl = blk_per_trk * heads
    p = Partition(self.rawblk, blk_num, len(self.parts), blk_per_cyl, self)
    p.read()
    self.parts.append(p)
    return True

  def change_partition(self, pid, drv_name=None, dev_flags=None, dos_type=None, flags=None, boot_pri=None, more_dos_env=None):
    # partition not found
    if pid < 0 or pid >= len(self.parts):
      return False
    p = self.parts[pid]
    pb = p.part_blk
    if pb == None:
      return False
    # set flags
    dirty = False
    if flags != None:
      pb.flags = flags
      dirty = True
    if drv_name != None:
      pb.drv_name = drv_name
      dirty = True
    if dev_flags != None:
      pb.dev_flags = dev_flags
      dirty = True
    # set dosenv flags
    if dos_type != None:
      pb.dos_env.dos_type = dos_type
      dirty = True
    if boot_pri != None:
      pb.dos_env.boot_pri = boot_pri
      dirty = True
    # update dos env
    if more_dos_env is not None:
      self._adjust_dos_env(pb.dos_env, more_dos_env)
      dirty = True
    # write change
    if dirty:
      pb.write()
    return True

  def delete_partition(self, pid):
    # partition not found
    if pid < 0 or pid >= len(self.parts):
      return False
    # unlink partition block
    next = Block.no_blk
    if (pid + 1) < len(self.parts):
      next = self.parts[pid+1].get_blk_num()
    if pid == 0:
      self.rdb.part_list = next
    else:
      last_pb = self.parts[-1]
      last_pb.part_blk.next = next
      last_pb.write()
    # free block
    p = self.parts[pid]
    blk_num = p.get_blk_num()
    self.used_blks.remove(blk_num)
    self._update_hi_blk()
    # write RDB
    self.rdb.write()
    # remove partition instance
    self.parts.remove(p)
    # relabel remaining parts
    num = 0
    for p in self.parts:
      p.num = num
      num += 1
    # done!
    return True

  # ----- file system handling ------

  def add_filesystem(self, data, dos_type=DosType.DOS1, version=0, dev_flags=None):
    # create a file system
    blk_num = self._next_rdb_block()
    fs_num = len(self.fs)
    fs = FileSystem(self.rawblk, blk_num, fs_num)
    # get total number of blocks for fs data
    num_blks = fs.get_total_blocks(data)
    # check if RDB has space left
    if not self._has_free_rdb_blocks(num_blks):
      return False
    # allocate blocks
    blks = self._alloc_rdb_blocks(num_blks)
    self.used_blks += blks
    self._update_hi_blk()
    # create file system
    fs.create(blks[1:], data, version, dos_type, dev_flags)
    fs.write()
    # link fs block
    if len(self.fs) == 0:
      # write into RDB
      self.rdb.fs_list = blk_num
    else:
      # write into last fs block
      last_fs = self.fs[-1]
      last_fs.fshd.next = blk_num
      last_fs.write(only_fshd=True)
    # update rdb: allocated blocks and optional link
    self.rdb.write()
    # add fs to list
    self.fs.append(fs)
    return True

  def delete_filesystem(self, fid):
    # check filesystem id
    if fid < 0 or fid >= len(self.fs):
      return False
    # unlink partition block
    next = Block.no_blk
    if (fid + 1) < len(self.fs):
      next = self.fs[fid+1].blk_num
    if fid == 0:
      self.rdb.fs_list = next
    else:
      last_fs = self.fs[-1]
      last_fs.fshd.next = next
      last_fs.write()
    # free block
    f = self.fs[fid]
    blks = f.get_blk_nums()
    for b in blks:
      self.used_blks.remove(b)
    self._update_hi_blk()
    # write RDB
    self.rdb.write()
    # remove partition instance
    self.fs.remove(f)
    # relabel remaining parts
    num = 0
    for f in self.fs:
      f.num = num
      num += 1
    # done!
    return True



import fsui
from launcher.ui.config.InputGroup import InputGroup
from launcher.ui.config.modelgroup import ModelGroup
from launcher.ui.removablemediagroup import RemovableMediaGroup
from launcher.ui.skin import Skin


class MainPanel(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        Skin.set_background_color(self)
        self.layout = fsui.VerticalLayout()
        self.model_group = ModelGroup(self)
        self.layout.add(self.model_group, fill=True)
        self.layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)
        self.removable_media_group = RemovableMediaGroup(self, 2, main=True)
        self.layout.add(self.removable_media_group, fill=True)
        self.layout.add_spacer(10)
        self.input_group = InputGroup(
            self, refresh_button=True, autofire_button=False
        )
        self.layout.add(self.input_group, fill=True)



from __future__ import absolute_import
from __future__ import print_function

from amitools.fs.block.BootBlock import BootBlock
from amitools.fs.block.RootBlock import RootBlock

from amitools.fs.validate.Log import Log
from amitools.fs.validate.BlockScan import BlockScan
from amitools.fs.validate.DirScan import DirScan
from amitools.fs.validate.FileScan import FileScan
from amitools.fs.validate.BitmapScan import BitmapScan
import amitools.fs.DosType as DosType

class Validator:
  """Validate an AmigaDOS file system"""
    
  def __init__(self, blkdev, min_level, debug=False, progress=None):
    self.log = Log(min_level)
    self.debug = debug
    self.blkdev = blkdev
    self.dos_type = None
    self.boot = None
    self.root = None
    self.block_scan = None
    self.progress = progress

  def scan_boot(self):
    """Step 1: scan boot block.
       Returns (True, x) if boot block has a valid dos type.
       Returns (x, True) if boot block is bootable
       Invalid checksum of the block is tolerated but remarked.
    """
    # check boot block
    boot = BootBlock(self.blkdev)
    boot.read()
    if boot.valid:
      # dos type is valid
      self.boot = boot
      self.dos_type = boot.dos_type
      # give a warning if checksum is not correct
      if not boot.valid_chksum:
        self.log.msg(Log.INFO,"invalid boot block checksum",0)
      self.log.msg(Log.INFO,"dos type is '%s'" % DosType.get_dos_type_str(self.dos_type))
      return (True, boot.valid_chksum)
    else:
      self.log.msg(Log.ERROR,"invalid boot block dos type",0)
      return (False, False)
  
  def scan_root(self):
    """Step 2: scan root block.
       Try to determine root block from boot block or guess number.
       Returns True if the root block could be decoded.
    """
    if self.boot != None:
      # retrieve root block number from boot block
      root_blk_num = self.boot.got_root_blk
      # check root block number
      if root_blk_num == 0:
        new_root = self.blkdev.num_blocks / 2
        self.log.msg(Log.INFO,"Boot contains not Root blk. Using default: %d" % new_root,root_blk_num)
        root_blk_num = new_root      
      elif root_blk_num < self.blkdev.reserved or root_blk_num > self.blkdev.num_blocks:
        new_root = self.blkdev.num_blocks / 2
        self.log.msg(Log.INFO,"Invalid root block number: given %d using guess %d" % (root_blk_num, new_root),root_blk_num)
        root_blk_num = new_root
    else:
      # guess root block number
      root_blk_num = self.blkdev.num_blocks / 2
      self.log.msg(Log.INFO,"Guessed root block number",root_blk_num)
    # read root block
    root = RootBlock(self.blkdev, root_blk_num)
    root.read()
    if not root.valid:
      self.log.msg(Log.INFO,"Root block is not valid -> No file system",root_blk_num)      
      self.root = None # mode without root
      return False
    else:
      self.root = root
      return True
  
  def scan_dir_tree(self):
    """Step 3: scan directory structure
       Return false if structure is not healthy"""
    self.block_scan = BlockScan(self.blkdev, self.log, self.dos_type)
    self.dir_scan = DirScan(self.block_scan, self.log)
    ok = self.dir_scan.scan_tree(self.root.blk_num, progress=self.progress)
    self.log.msg(Log.INFO,"Scanned %d directories" % len(self.dir_scan.get_all_dir_infos()))
    if self.debug:
      self.dir_scan.dump()
          
  def scan_files(self):
    """Step 4: scan through all found files"""
    self.file_scan = FileScan(self.block_scan, self.log, self.dos_type)
    all_files = self.dir_scan.get_all_file_hdr_blk_infos()
    self.log.msg(Log.INFO,"Scanning %d files" % len(all_files))
    self.file_scan.scan_all_files(all_files, progress=self.progress)
    if self.debug:
      self.file_scan.dump()
      
  def scan_bitmap(self):
    """Step 5: validate block bitmap"""
    self.bitmap_scan = BitmapScan(self.block_scan, self.log)
    self.bitmap_scan.scan_bitmap(self.root)

  def get_summary(self):
    """Return (errors, warnings) of log"""
    num_errors = self.log.get_num_level(Log.ERROR)
    num_warns = self.log.get_num_level(Log.WARN)
    return (num_errors, num_warns)



from fsui.qt import QPixmap, QImage, QIcon
from fsbc.resources import Resources
from .Image import Image


class Icon(object):
    """
    FIXME: Add support for icon search path, so, during startup, a list
    of package data directories to check can be added to this class, and
    icons can be looked up just by name.
    """

    def __init__(self, name, path):
        self.name = name
        self.path = path

    def stream_for_size(self, size):
        # package, file_ = name.split(":", 1)
        assert self.path.startswith("pkg:")
        package = self.path[4:]
        try:
            name = "res/{0}x{0}/{1}.png".format(size, self.name)
            stream = Resources(package).stream(name)
        except LookupError:
            name = "res/{0}/{1}.png".format(size, self.name)
            stream = Resources(package).stream(name)
        return stream

    def qimage(self, size):
        stream = self.stream_for_size(size)
        qimage = QImage()
        qimage.loadFromData(stream.read())
        return qimage

    def image(self, size):
        return Image(object=self.qimage(size))

    def qpixmap(self, size):
        return QPixmap(self.qimage(size))

    def qicon(self, size=None):
        icon = QIcon()
        if size is not None:
            sizes = [size]
        else:
            sizes = [16, 22, 24, 32, 40, 48, 64, 128, 256]
        for size in sizes:
            try:
                pixmap = self.qpixmap(size)
            except Exception:
                pass
            else:
                icon.addPixmap(pixmap)
        return icon



# FSGS - Common functionality for FS Game System.
# Copyright (C) 2013-2016  Frode Solheim <frode@openretro.org>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
FSGS Game driver for Atari ST family.

TODO:

* Fullscreen stretching with SDL2 version.
* Multi-monitor awareness with SDL2 version?
* Support saving disk changes.
* Support (persistent) save states.
* Check and/or fix support for vsync.
* NTSC vs PAL?.
* Proper joystick support and keyboard joystick emulation.
* Mouse in port 0.
* Multiple floppies and floppy swapping.
* Screenshots are saved to the (temp) current working directory, not the
  global screenshots directory.

"""

import os
import traceback

from fsgs.drivers.gamedriver import GameDriver
from fsgs.knownfiles import KnownFile
from fsgs.option import Option
from fsgs.platforms import PLATFORM_ATARI

ATARI_MODEL_520ST = "520st"
ATARI_MODEL_1040ST = "1040st"
ATARI_WIDTH = 832
# ATARI_HEIGHT = 576
ATARI_HEIGHT = 552
# noinspection SpellCheckingInspection
ATARI_TOS_102_UK = KnownFile(
    "87900a40a890fdf03bd08be6c60cc645855cbce5",
    PLATFORM_ATARI,
    "TOS v1.02 (1987)(Atari Corp)(Mega ST)(GB)[MEGA TOS].img",
)
# noinspection SpellCheckingInspection
ATARI_TOS_104_UK = KnownFile(
    "9526ef63b9cb1d2a7109e278547ae78a5c1db6c6",
    PLATFORM_ATARI,
    "TOS v1.04 (1989)(Atari Corp)(Mega ST)(GB)[Rainbow TOS].img",
)
# noinspection SpellCheckingInspection
ATARI_TOS_1062_UK = KnownFile(
    "70db24a7c252392755849f78940a41bfaebace71",
    PLATFORM_ATARI,
    "TOS v1.62 (1990)(Atari)(GB)[STE TOS, Rev 2][STE].img",
)


class HatariAtariDriver(GameDriver):
    def __init__(self, fsgs):
        super().__init__(fsgs)
        self.emulator.name = "hatari-fs"
        self.floppies_dir = self.temp_dir("Media")
        self.tos_file = self.temp_file("tos.img")
        self.config_file = self.temp_file("hatari.cfg")
        self.helper = AtariHelper(self.config)
        self.floppies = []

    def prepare(self):
        super().prepare()
        self.prepare_tos()
        self.prepare_floppies()
        with open(self.config_file.path, "w", encoding="UTF-8") as f:
            self.write_config_file(f)
        self.emulator.args.extend(["--configfile", self.config_file.path])

    def install(self):
        super().install()

    def run(self):
        super().run()

    def finish(self):
        super().finish()

    def prepare_floppies(self):
        original_floppies = []
        if self.config["floppy_drive_0"]:
            original_floppies.append(self.config["floppy_drive_0"])
        floppies = []
        for p in original_floppies:
            dest_path = os.path.join(
                self.floppies_dir.path, os.path.basename(p)
            )
            self.files.add(dest_path, source=p)
            floppies.append(dest_path)
        self.floppies = floppies

    def prepare_tos(self):
        tos_file = self.helper.tos_file()
        self.files.add(
            self.tos_file.path, sha1=tos_file.sha1, description=tos_file.name
        )

    def write_config_file(self, f):
        f.write("[Screen]\n")
        # f.write("nMonitorType = 1\n")
        # f.write("nFrameSkips = 5\n")
        if self.use_fullscreen():
            f.write("bFullScreen = TRUE\n")
        else:
            f.write("bFullScreen = FALSE\n")
        # f.write("bKeepResolution = TRUE\n")
        f.write("bAllowOverscan = FALSE\n")
        # f.write("nSpec512Threshold = 1\n")
        # f.write("nForceBpp = 0\n")
        # f.write("bAspectCorrect = TRUE\n")
        # f.write("bAspectCorrect = FALSE\n")
        # f.write("bUseExtVdiResolutions = FALSE\n")
        # f.write("nVdiWidth = 640\n")
        # f.write("nVdiHeight = 480\n")
        # f.write("nVdiColors = 2\n")
        # f.write("bMouseWarp = TRUE\n")
        f.write("bShowStatusbar = FALSE\n")
        f.write("bShowDriveLed = FALSE\n")
        # f.write("bCrop = FALSE\n")
        # f.write("bForceMax = FALSE\n")
        # f.write("nMaxWidth = 832\n")
        # f.write("nMaxHeight = 588\n")
        f.write("nRenderScaleQuality = 1\n")
        f.write("bUseVsync = 0\n")

        f.write("[Sound]\n")
        if self.use_audio_frequency():
            f.write("nPlaybackFreq = {0}\n".format(self.use_audio_frequency()))

        f.write("[Floppy]\n")
        num_floppy_drives = 2
        inserted_floppies = self.floppies[:num_floppy_drives]
        if len(inserted_floppies) > 0:
            f.write(
                "szDiskAFileName = {path}\n".format(path=inserted_floppies[0])
            )
        if len(inserted_floppies) > 1:
            f.write(
                "szDiskBFileName = {path}\n".format(path=inserted_floppies[1])
            )

        f.write("[ROM]\n")
        f.write("szTosImageFileName = {}\n".format(self.tos_file.path))
        f.write("bPatchTOS = FALSE\n")

        f.write("[System]\n")
        if self.helper.model_family() == "ST":
            f.write("nModelType = 0\n")
        elif self.helper.model_family() == "STE":
            f.write("nModelType = 2\n")
        else:
            print("WARNING: UNKNOWN ST MODEL FAMILY")
        accuracy = self.helper.accuracy()
        if accuracy == 1:
            f.write("bCycleExactCpu = TRUE\n")
            f.write("bCompatibleCpu = TRUE\n")
            f.write("bPatchTimerD = FALSE\n")
        elif accuracy == 0:
            f.write("bCycleExactCpu = FALSE\n")
            f.write("bCompatibleCpu = TRUE\n")
            f.write("bPatchTimerD = TRUE\n")
        else:
            f.write("bCycleExactCpu = FALSE\n")
            f.write("bCompatibleCpu = FALSE\n")
            f.write("bPatchTimerD = TRUE\n")
        f.write("bFastBoot = FALSE\n")
        f.write("bBlitter = FALSE\n")

        if self.use_stretching():
            self.env["FSGS_STRETCH"] = "1"
        else:
            self.env["FSGS_STRETCH"] = "0"

        f.write("[Memory]\n")
        memory = self.helper.memory()
        if memory == 1024:
            f.write("nMemorySize = 1\n")
        elif memory == 512:
            f.write("nMemorySize = 0\n")
        else:
            print("WARNING: UNKNOWN ST MEMORY SIZE")

    def configure_fullscreen_scaling(self):
        """Configure scaling for the SDL 1.x version of Hatari-FS.

        Not (currently) used with the SDL 2.x version of Hatari-FS.
        """
        sx, sy, sw, sh = 0, 0, ATARI_WIDTH, ATARI_HEIGHT
        try:
            viewport = self.config["viewport"]
            if viewport:
                sx, sy, sw, sh = viewport.rsplit("=", 1)[-1].strip().split(" ")
                sx, sy, sw, sh = int(sx), int(sy), int(sw), int(sh)
        except Exception:
            traceback.print_exc("Could not get viewport information")
        screen_w, screen_h = self.screen_size()
        print("viewport is", sx, sy, sw, sh)
        if self.use_stretching():
            target_w = screen_w
            target_h = screen_h
        else:
            scale = min(screen_w / sw, screen_h / sh)
            target_w = sw * scale
            target_h = sh * scale
        scale_x = target_w / sw
        scale_y = target_h / sh
        print("scale", scale_x, scale_y)
        orig_scale_x = min(screen_w / ATARI_WIDTH, screen_h / ATARI_HEIGHT)
        orig_scale_y = orig_scale_x
        print("org_scale = ", orig_scale_x, orig_scale_y)
        offset_x = -(sx + (sw / 2.0) - ATARI_WIDTH / 2.0) / (ATARI_WIDTH / 2.0)
        offset_y = -(sy + (sh / 2.0) - ATARI_HEIGHT / 2.0) / (
            ATARI_HEIGHT / 2.0
        )
        hz = scale_x / orig_scale_x
        vz = scale_y / orig_scale_y
        print("horizontal zoom:", hz / 100000000.0)
        print("vertical zoom:", vz / 100000000.0)
        self.env["FILTER_VERT_OFFSET"] = str(offset_y)
        self.env["FILTER_HORIZ_OFFSET"] = str(offset_x)
        self.env["FILTER_VERT_ZOOM"] = str(vz)
        self.env["FILTER_HORIZ_ZOOM"] = str(hz)


class AtariHelper:
    def __init__(self, config):
        self.config = config

    def accuracy(self):
        try:
            accuracy = int(self.config.get(Option.ACCURACY, "1"))
        except ValueError:
            accuracy = 1
        return accuracy

    def model(self):
        model = self.config.get(Option.ATARI_MODEL)
        if not model:
            model = ATARI_MODEL_520ST
        return model

    def model_family(self):
        model = self.config.get(Option.ATARI_MODEL, "").upper()
        if model.endswith("ST"):
            return "ST"
        elif model.endswith("STE"):
            return "STE"
        return ""

    def tos_file(self):
        model = self.model().upper()
        if "STE" in model:
            return ATARI_TOS_1062_UK
        else:
            return ATARI_TOS_102_UK

    def memory(self):
        model = self.model().upper()
        if model.startswith("1040"):
            return 1024
        else:
            return 512



from fsgs.ogd.refresh import DatabaseRefreshTask
import fsui
from fsui.extra.iconheader import IconHeader

# from workspace.shell import SimpleApplication
from launcher.res import gettext
from launcher.ui.widgets import CloseButton
from workspace.ui.theme import WorkspaceTheme


class RefreshWindow(fsui.Window):
    @classmethod
    def open(cls, parent=None):
        return fsui.open_window_instance(cls, parent)

    def __init__(self, parent=None):
        print("RefreshWindow parent =", parent)
        title = gettext("Updating Database")
        super().__init__(parent, title, maximizable=False)
        self.set_icon(fsui.Icon("refresh", "pkg:workspace"))
        self.theme = WorkspaceTheme.instance()
        self.layout = fsui.VerticalLayout()

        self.layout.min_width = 500
        self.layout.set_padding(20, 20, 20, 20)

        self.icon_header = IconHeader(
            self,
            fsui.Icon("refresh", "pkg:workspace"),
            gettext("Updating Database"),
            "",
        )
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        self.created_label = fsui.Label(self, "")
        hori_layout.add(self.created_label, expand=True)
        hori_layout.add_spacer(20)
        self.stop_button = fsui.Button(self, gettext("Stop"))
        self.stop_button.activated.connect(self.on_abort_activated)
        hori_layout.add(self.stop_button)

        if self.window.theme.has_close_buttons:
            self.close_button = CloseButton(self)
            hori_layout.add(self.close_button, fill=True, margin_left=10)

        self.set_size(self.layout.get_min_size())
        self.center_on_parent()

        self.task = DatabaseRefreshTask()
        self.task.progressed.connect(self.on_progress)
        self.task.failed.connect(self.on_failure)
        self.task.succeeded.connect(self.close)
        self.task.stopped.connect(self.close)
        self.task.start()

        # self.closed.connect(self.on_closed_signal)

    def __del__(self):
        print("RefreshWindow.__del__")

    def on_close(self):
        print("RefreshWindow.on_close")
        self.task.stop()

    # def on_closed_signal(self):
    #     self.task.stop()

    def on_abort_activated(self):
        self.task.stop()
        self.stop_button.disable()

    def on_failure(self, message):
        fsui.show_error(message, parent=self.get_window())

    def on_progress(self, message):
        if not isinstance(message, str):
            message = message[0]
        # print("on_progress", status)
        self.icon_header.subtitle_label.set_text(message)


# application = SimpleApplication(RefreshWindow)



from __future__ import absolute_import
from __future__ import print_function

from .BlockDevice import BlockDevice
import os.path
import os

class PartBlockDevice(BlockDevice):
  def __init__(self, raw_blkdev, part_blk, auto_close=False):
    self.raw_blkdev = raw_blkdev
    self.part_blk = part_blk
    self.blk_off = 0
    self.auto_close = auto_close

  def open(self):
    # extract geometry from partition block
    dos_env = self.part_blk.dos_env
    lo_cyl = dos_env.low_cyl
    hi_cyl = dos_env.high_cyl
    cyls = hi_cyl - lo_cyl + 1
    heads = dos_env.surfaces
    secs = dos_env.blk_per_trk
    block_bytes = dos_env.block_size * 4
    reserved = dos_env.reserved
    boot_blocks = dos_env.boot_blocks
    if boot_blocks == 0:
      boot_blocks = 2
    self._set_geometry(cyls, heads, secs, block_bytes, reserved, boot_blocks)
    # calc block offset of partition
    self.blk_off = heads * secs * lo_cyl
    return True

  def flush(self):
    self.raw_blkdev.flush()

  def close(self):
    # auto close containing rdisk
    if self.auto_close:
      self.raw_blkdev.close()

  def read_block(self, blk_num):
    if blk_num >= self.num_blocks:
      raise ValueError("Invalid Part block num: got %d but max is %d" % (blk_num, self.num_blocks))
    return self.raw_blkdev.read_block(self.blk_off + blk_num)

  def write_block(self, blk_num, data):
    if blk_num >= self.num_blocks:
      raise ValueError("Invalid Part block num: got %d but max is %d" % (blk_num, self.num_blocks))
    if len(data) != self.block_bytes:
      raise ValueError("Invalid Part block size written: got %d but size is %d" % (len(data), self.block_bytes))
    self.raw_blkdev.write_block(self.blk_off + blk_num, data)



import nose.tools

import fsbc.signal


def test_mypy():
    from nose.plugins.skip import SkipTest

    raise SkipTest()
    # import fstd.mypy
    # fstd.mypy.check_module(fsbc.signal.__name__)


def test_doctest():
    import doctest

    failure_count, test_count = doctest.testmod(fsbc.signal)
    nose.tools.assert_equals(failure_count, 0)
